# Copyright 2022 Cochise Ruhulessin
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from typing import Any
from typing import Callable
from typing import Generator

from ..cryptographicfailure import UnknownAlgorithm


class Algorithm(str):
    """The base class for all cryptographic algorithms."""
    __module__: str = 'ckms.algorithm'
    _aliases: dict[str, str] = {}
    _defaults: dict[tuple[str, str], 'Algorithm'] = {}
    _registry: dict[str, 'Algorithm'] = {}
    allowed_ops: list[str]
    default_ops: list[str]
    digest: str | None
    digest_oid: str | None
    padding: str | None
    kty: str
    use: str

    @classmethod
    def alias(cls, name: str, alias_of: str) -> None:
        cls._aliases[name] = alias_of

    @classmethod
    def default(cls, kty: str, use: str) -> 'Algorithm':
        return Algorithm._defaults[(kty, use)]

    @classmethod
    def get(cls, name: str) -> 'Algorithm':
        if name in cls._aliases:
            name = cls._aliases[name]
        return cls._registry[name]

    @classmethod
    def get_for_curve(cls, curve: str) -> 'Algorithm':
        # TODO:
        curves = {
            'P-256': 'ES256',
            'P-384': 'ES384',
            'P-521': 'ES512',
            'P-256K': 'ES256K',
        }
        return cls.get(curves[curve])

    @staticmethod
    def register(
        name: str,
        *,
        base: type['Algorithm'],
        **params: Any
    ) -> None:
        """Register a new algorithm preset using the given class."""
        Algorithm._registry[name] = base(name, **params)

    @classmethod
    def __get_validators__(
        cls
    ) -> Generator[Callable[..., 'Algorithm'], None, None]:
        yield cls.validate

    @classmethod
    def validate(cls, value: str) -> 'Algorithm':
        if str(value) not in Algorithm._registry:
            raise UnknownAlgorithm(value)
        if isinstance(value, cls): # pragma: no cover
            return value
        return Algorithm._registry[value]

    def __new__(
        cls,
        name: str,
        allowed_ops: set[str] | None = None,
        default_ops: set[str] | None = None,
        digest: str | None = None,
        digest_oid: str | None = None,
        padding: str | None = None,
        kty: str | None = None,
        use: str | None = None,
        crv: str | None = None,
        default: bool = False,
        *args: Any,
        **kwargs: Any
    ):
        if name in Algorithm._aliases:
            name = Algorithm._aliases[name]
        if name in Algorithm._registry:
            return Algorithm._registry[name]

        self = str.__new__(cls, name)
        self.allowed_ops = allowed_ops or set()
        self.default_ops = default_ops or set()
        self.digest = digest
        self.digest_oid = digest_oid
        self.padding = padding
        self.use = use # type: ignore
        if kty is not None:
            self.kty = kty
        if kty and use and default:
            Algorithm._defaults[(kty, use)] = self
        return self

    def get_digest_oid(self) -> str | None:
        """Return the Object Identifier (OID) for this algorithm when
        used to create a digital signature.
        """
        return self.digest_oid

    def get_key_ops(self, public: bool) -> list[str]:
        """Return the default allowed key operations for this algorithm."""
        if not public:
            return self.default_ops
        ops: list[str] = []
        if 'decrypt' in self.default_ops:
            ops.append('encrypt')
        if 'sign' in self.default_ops:
            ops.append('verify')
        if 'unwrapKey' in self.default_ops:
            ops.append('wrapKey')
        return ops


    def get_hasher(self) -> Any:
        """Return the hasher used to perform a cryptographic operation."""
        raise NotImplementedError

    def get_padding(self) -> Any:
        """Return an instance of the padding, if the algorithm implements it."""
        raise NotImplementedError


# Copyright 2018 Cochise Ruhulessin
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Declares :class:`RSAAlgorithmType`."""
import enum


class RSAAlgorithmType(str, enum.Enum):
    RS256       = 'RS256'
    RS384       = 'RS384'
    RS512       = 'RS512'
    PS256       = 'PS256'
    PS384       = 'PS384'
    PS512       = 'PS512'
    RSA1_5      = 'RSA1_5'
    RSA_OAEP    = 'RSA-OAEP'
    RSA_OAEP256 = 'RSA-OAEP-256'
    RSA_OAEP384 = 'RSA-OAEP-384'
    RSA_OAEP512 = 'RSA-OAEP-512'


# Copyright 2022 Cochise Ruhulessin
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# type: ignore
import copy
import json
from typing import Any

import pytest
from jwcrypto.jwe import default_allowed_algs as default_jwe_algs
from jwcrypto.jwe import JWE
from jwcrypto.jwk import JWK
from jwcrypto.jws import default_allowed_algs as default_jws_algs
from jwcrypto.jws import JWS
from jwcrypto.jws import InvalidJWSSignature

from ckms.core.const import SIGNING_ALGORITHMS
from ckms.core.models import KeySpecification
from ckms.core.provider.tests import JOSEProducerTestCase
from ckms.core.provider.tests import JOSEEncryptionTestCase
from ckms.jose.decoder import Decoder
from ckms.jose.encoder import Encoder
from ckms.utils import b64encode


ENCRYPTION_ALGORITHMS: list[str] = list(
    sorted(set(default_jwe_algs) - {'dir'})
)


class TestEncryptionCompatibility(JOSEEncryptionTestCase):
    supported_wrapping_algorithms: set[str] = set(ENCRYPTION_ALGORITHMS)

    def jwk_from_spec(
        self,
        spec: KeySpecification,
        private: bool = False
    ) -> dict[str, Any]:
        return super().jwk_from_spec(spec, True)

    async def decrypt_jwe(
        self,
        encoder: Encoder,
        decoder: Decoder,
        recipients: list[tuple[KeySpecification, dict[str, Any] | None]],
        signers: list[tuple[KeySpecification, dict[str, Any] | None, dict[str, Any] | None]],
        payload: bytes | dict[str, Any],
        token: str | dict[str, Any]
    ) -> Any:
        jwe = JWE()
        jwe.deserialize(token)
        for spec, header in recipients:
            jwk = JWK(**self.jwk_from_spec(spec))
            jwe.decrypt(jwk)
            break
        return jwe.payload


@pytest.mark.parametrize("name", default_jws_algs)
class TestDeserializeCompatibility(JOSEProducerTestCase):

    async def verify_jws(
        self,
        verifiers: KeySpecification,
        token: bytes,
        payload: str | None = None,
        expect_jwt: bool = False
    ):
        jws = self.deserialize_jws(token)
        if payload is not None:
            assert jws.objects.get('payload') == str.encode(payload, 'utf-8')
        verified = False
        for spec in verifiers:
            k = JWK(**self.jwk_from_spec(spec))
            try:
                jws.verify(k)
            except InvalidJWSSignature:
                break
        else:
            verified = True
        return verified

    def deserialize_jws(self, token: str) -> JWS:
        jws = JWS()
        jws.deserialize(token) # type: ignore
        return jws


# pylint: skip-file
# Copyright 2018 Cochise Ruhulessin
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from typing import Any

import pydantic
from cryptography.hazmat.primitives.serialization import load_pem_public_key
from google.cloud.kms import KeyManagementServiceAsyncClient

from ckms.core.models import KeySpecification
from ckms.types import Algorithm
from ckms.types import CryptographyPublicKeyType
from ckms.types import IKeyInspector
from ckms.types import IProvider
from ckms.types import AESAlgorithmType
from ckms.types import EllipticCurveAlgorithmType
from ckms.types import HMACAlgorithmType
from ckms.types import KeyAlgorithmType
from ckms.types import KeyUseType
from ckms.types import RSAAlgorithmType
from .igoogleprovider import IGoogleProvider


ALGORITHM_MAP: dict[str, tuple[KeyAlgorithmType, str, KeyUseType]] = {
    'EC_SIGN_SECP256K1_SHA256'      : (EllipticCurveAlgorithmType.ES256K, 'EC', KeyUseType.sign),
    'EC_SIGN_P256_SHA256'           : (EllipticCurveAlgorithmType.ES256, 'EC', KeyUseType.sign),
    'EC_SIGN_P384_SHA384'           : (EllipticCurveAlgorithmType.ES384, 'EC', KeyUseType.sign),
    'RSA_DECRYPT_OAEP_2048_SHA1'    : (RSAAlgorithmType.RSA_OAEP, 'RSA', KeyUseType.encrypt),
    'RSA_DECRYPT_OAEP_2048_SHA256'  : (RSAAlgorithmType.RSA_OAEP256, 'RSA', KeyUseType.encrypt),
    'RSA_DECRYPT_OAEP_3072_SHA1'    : (RSAAlgorithmType.RSA_OAEP, 'RSA', KeyUseType.encrypt),
    'RSA_DECRYPT_OAEP_3072_SHA256'  : (RSAAlgorithmType.RSA_OAEP256, 'RSA', KeyUseType.encrypt),
    'RSA_DECRYPT_OAEP_4096_SHA1'    : (RSAAlgorithmType.RSA_OAEP, 'RSA', KeyUseType.encrypt),
    'RSA_DECRYPT_OAEP_4096_SHA256'  : (RSAAlgorithmType.RSA_OAEP256, 'RSA', KeyUseType.encrypt),
    'RSA_DECRYPT_OAEP_4096_SHA512'  : (RSAAlgorithmType.RSA_OAEP512, 'RSA', KeyUseType.encrypt),
    'RSA_SIGN_PSS_2048_SHA256'      : (RSAAlgorithmType.PS256, 'RSA', KeyUseType.sign),
    'RSA_SIGN_PSS_3072_SHA256'      : (RSAAlgorithmType.PS256, 'RSA', KeyUseType.sign),
    'RSA_SIGN_PSS_4096_SHA256'      : (RSAAlgorithmType.PS256, 'RSA', KeyUseType.sign),
    'RSA_SIGN_PSS_4096_SHA512'      : (RSAAlgorithmType.PS512, 'RSA', KeyUseType.sign),
    'RSA_SIGN_PKCS1_2048_SHA256'    : (RSAAlgorithmType.RS256, 'RSA', KeyUseType.sign),
    'RSA_SIGN_PKCS1_3072_SHA256'    : (RSAAlgorithmType.RS256, 'RSA', KeyUseType.sign),
    'RSA_SIGN_PKCS1_4096_SHA256'    : (RSAAlgorithmType.RS256, 'RSA', KeyUseType.sign),
    'RSA_SIGN_PKCS1_4096_SHA512'    : (RSAAlgorithmType.RS512, 'RSA', KeyUseType.sign),
    'HMAC_SHA256'                   : (HMACAlgorithmType.HS256, 'oct', KeyUseType.sign),
    'GOOGLE_SYMMETRIC_ENCRYPTION'   : (AESAlgorithmType.A256GCM, 'oct', KeyUseType.encrypt),
}


class GoogleKeySpecification(KeySpecification):
    kty: str = 'none'
    algorithm: Algorithm | None
    use: KeyUseType | None
    project: str | None
    location: str
    keyring: str = pydantic.Field(..., alias='key_ring')
    key: str = pydantic.Field(..., alias='crypto_key')
    version: int
    public: CryptographyPublicKeyType | None

    @property
    def crypto_key(self) -> str:
        return KeyManagementServiceAsyncClient.crypto_key_path(
            project=self.project,
            location=self.location,
            key_ring=self.keyring,
            crypto_key=self.key
        )

    @property
    def resource_name(self) -> str:
        return KeyManagementServiceAsyncClient.crypto_key_version_path(
            project=self.project,
            location=self.location,
            key_ring=self.keyring,
            crypto_key=self.key,
            crypto_key_version=self.version
        )

    @classmethod
    def autodiscover(
        cls,
        provider: IProvider,
        inspector: IKeyInspector,
        values: dict[str, Any]
    ) -> None:
        if not values.get('project'): # pragma: no cover
            values['project'] = provider.project # type: ignore

    async def load(self) -> 'GoogleKeySpecification':
        provider: IGoogleProvider = self.provider # type: ignore
        inspector: IKeyInspector = provider.inspector
        async with provider.client() as client:
            name = self.resource_name
            version = await client.get_crypto_key_version( # type: ignore
                request={'name': name},
                timeout=5.0
            )
            if version.algorithm.name in ALGORITHM_MAP:
                algorithm, self.kty, self.use = ALGORITHM_MAP[version.algorithm.name] # type: ignore
                self.algorithm = Algorithm.get(algorithm.value) # type: ignore
                self.allow = {self.algorithm}
            else:
                raise NotImplementedError(version)
            self.curve = inspector.get_algorithm_curve(self.algorithm)

            # Normally, the kid is generated from the k value of the JWK,
            # but we have no access so use the resource name in Google
            # Cloud KMS.
            if self.is_symmetric():
                self.kid = provider.get_key_identifier(name)

            # Fetch public key for asymmetric types.
            if self.is_asymmetric():
                public = await client.get_public_key( # type: ignore
                    request={'name': name},
                    timeout=5.0
                )
                self.public = load_pem_public_key(public.pem.encode()) # type: ignore
                self.kid = self.provider.inspector.calculate_kid(self.public)

        self.loaded = True
        return self

    def get_public_key(self) -> CryptographyPublicKeyType | None:
        return self.public

    def has_key_material(self) -> bool:
        return False

    def __hash__(self) -> int: # type: ignore
        return self.resource_name.__hash__()


import asyncio

import ckms.core


async def main():
    # Generate an RSA key, marked for use with the 'PS256' algorithm, using
    # the 'local' cryptographic operations provider - an implementation
    # that performs the operations using in-memory key material.
    spec = ckms.core.parse_spec({
        'provider': 'local',
        'kty': 'RSA',
        'algorithm': 'PS256',
        'key': {'length': 2048}
    })

    # Before using a key, it must be configured once by awaiting it.
    # Alternatively, the ckms.core.parse_spec() call may be awaited.
    await spec

    data = b'Hello world!'
    sig = await spec.sign(data)
    assert await spec.verify(sig, data)
    assert not await spec.verify(sig, b'Not the signed data')


if __name__ == '__main__':
    asyncio.run(main())



# Copyright 2022 Cochise Ruhulessin
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import pytest

from ckms.types import Message
from ckms.core.models import KeySpecification
from .baseprovidertestcase import BaseProviderTestCase


class AsymmetricSigningTestCase(BaseProviderTestCase):

    @pytest.mark.asyncio
    async def test_spec_to_jwk(self, spec: KeySpecification):
        await spec
        jwk = spec.as_jwk()
        assert jwk.kty == spec.kty
        assert jwk.alg == spec.algorithm

    @pytest.mark.asyncio
    async def test_verify_from_jwk(
        self,
        spec: KeySpecification,
        data: bytes
    ):
        await spec
        jwk = spec.as_jwk(private=False)
        sig = await self.sign(spec, data)
        await self.verify(jwk, sig, data)

    @pytest.mark.asyncio
    async def test_sign_bytes(
        self,
        spec: KeySpecification,
        data: bytes
    ) -> None:
        if not spec.can_sign():
            pytest.skip()
        await self.sign(spec, data)

    @pytest.mark.asyncio
    async def test_sign_message(
        self,
        spec: KeySpecification,
        data: bytes
    ) -> None:
        if not spec.can_sign():
            pytest.skip()
        await self.sign(spec, Message(buf=data))

    @pytest.mark.asyncio
    async def test_verify_message(
        self,
        spec: KeySpecification,
        data: bytes
    ) -> None:
        if not spec.can_sign():
            pytest.skip()
        sig = await self.sign(spec, Message(buf=data))
        assert await self.verify(spec.as_public(), sig, Message(buf=data))

    @pytest.mark.asyncio
    async def test_asymmetric_verify_from_provider(
        self,
        spec: KeySpecification,
        data: bytes
    ) -> None:
        if not spec.is_asymmetric() or not spec.can_sign():
            pytest.skip()
        sig = await self.sign(spec, data)
        assert await self.verify(spec.as_public(), sig, data)

    @pytest.mark.asyncio
    async def test_asymmetric_verify_from_provider_invalid(
        self,
        spec: KeySpecification,
        data: bytes
    ) -> None:
        if not spec.is_asymmetric() or not spec.can_sign():
            pytest.skip()
        sig = await self.sign(spec, data)
        assert not await self.verify(spec.as_public(), sig, b'foo')

    @pytest.mark.asyncio
    async def test_asymmetric_verify_from_provider_signature_too_short(
        self,
        spec: KeySpecification,
        data: bytes
    ) -> None:
        if not spec.is_asymmetric() or not spec.can_sign():
            pytest.skip()
        sig = await self.sign(spec, data)
        assert not await self.verify(spec.as_public(), sig[:-1], data)

    @pytest.mark.asyncio
    async def test_asymmetric_verify_from_provider_signature_too_long(
        self,
        spec: KeySpecification,
        data: bytes
    ) -> None:
        if not spec.is_asymmetric() or not spec.can_sign():
            pytest.skip()
        sig = await self.sign(spec, data)
        assert not await self.verify(spec.as_public(), sig + b'0', data)


# Copyright 2022 Cochise Ruhulessin
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# type: ignore
from typing import Any

import dsnparse

from .keychain import get_default_keychain
from .keychain import Keychain
from .keyinspector import KeyInspector
from .provider import Provider
from .models import RemoteBlobParams as RemoteBlob
from .models import KeySpecification


__all__ = [
   'get_default_keychain',
   'types',
   'KeyInspector',
   'KeySpecification',
   'Provider',
   'RemoteBlob',
]


def parse_dsn(dsn: str | dsnparse.ParseResult) -> dict[str, Any]:
   """Parse a key specification from a Data Source Name (DSN)."""
   if isinstance(dsn, str):
      dsn = dsnparse.parse(dsn)
   p = provider(dsn.scheme)
   return p.parse_dsn(dsn)


def parse_spec(spec: dict[str, Any]) -> KeySpecification:
    """Parse the specification of a *key* using the parameters defined by the
    specification `spec`.
    """
    p = provider(spec.get('provider'))
    return p.parse_spec(spec)


def provider(name: str) -> Provider:
    """Configure the provider with the given name and return the
    configured instance.
    """
    return Provider.get(name) # type: ignore


async def fetch(spec: RemoteBlob | str, **params: Any) -> bytes:
   """Fetches the content of the remote blob using the parameters
   specified. Return a byte-string holding the content.
   """
   if isinstance(spec, str):
      spec = RemoteBlob(provider=spec, **params)
   provider = Provider.get(spec.provider)
   return await provider.fetch(spec) # type: ignore



# Copyright 2022 Cochise Ruhulessin
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from typing import Awaitable
from typing import Any

from .digest import Digest
from .ioperationperformer import IOperationPerformer
from .iprovider import IProvider
from .keyoperationtype import KeyOperationType
from .message import Message


class Verifier(IOperationPerformer):
    __module__: str = 'ckms.types'
    algorithm: Any
    provider: IProvider

    @IOperationPerformer.must_allow(KeyOperationType.verify)
    def verify(
        self,
        signature: bytes,
        message: bytes | Digest | Message,
    ) -> bool | Awaitable[bool]:
        """Verify that `signature` was created using this key from the given
        `message`.
        """
        assert isinstance(self.provider, IProvider)
        assert self.algorithm is not None
        digest = self.provider.get_hashing_algorithm(self.algorithm)
        if not isinstance(message, (Digest, Message)):
            message = Message(buf=message)
        if digest is not None and not message.digestmod:
            message.set_digest_algorithm(digest)
        return self.provider.verify(
            algorithm=self.algorithm,
            key=self,
            message=message,
            signature=signature
        )


# pylint: skip-file
# type: ignore
import json

import jwcrypto.jwe
import jwcrypto.jwk
import pydantic
import pytest
from jwcrypto.common import json_encode

from ckms.core.models import JSONWebEncryption
from ckms.core.models import JSONWebEncryptionFlattened
from ckms.core.models import JSONWebEncryptionWithRecipients
from ckms.utils import b64encode


@pytest.fixture
def token():
    payload = b'Hello world!'
    k1 = jwcrypto.jwk.JWK.generate(kty='oct', size=256)
    k2 = jwcrypto.jwk.JWK.generate(kty='oct', size=256)
    jwe = jwcrypto.jwe.JWE(
        payload,
        json_encode({
            'alg': "A256KW",
            'enc': "A256CBC-HS512"
        })
    )
    jwe.add_recipient(k1, {'foo': 1})
    jwe.add_recipient(k2, {'bar': 2})
    return jwe


@pytest.fixture
def token_single():
    payload = b'Hello world!'
    k1 = jwcrypto.jwk.JWK.generate(kty='oct', size=256)
    jwe = jwcrypto.jwe.JWE(
        payload,
        json_encode({
            'alg': "A256KW",
            'enc': "A256CBC-HS512",
            'bar': 2
        })
    )
    jwe.add_recipient(k1, {'foo': 1})
    return jwe


def test_parse_jwe_multi_recipient(token: jwcrypto.jwe.JWE):
    obj = JSONWebEncryption.parse_obj(json.loads(token.serialize()))
    assert isinstance(obj.__root__, JSONWebEncryptionWithRecipients)
    assert len(obj.recipients) == 2


def test_parse_jwe_single_recipient():
    payload = b'Hello world!'
    k1 = jwcrypto.jwk.JWK.generate(kty='oct', size=256)
    jwe = jwcrypto.jwe.JWE(
        payload,
        json_encode({
            'alg': "A256KW",
            'enc': "A256CBC-HS512"
        })
    )
    jwe.add_recipient(k1)
    obj = JSONWebEncryption.parse_obj(json.loads(jwe.serialize()))
    header = obj.header
    assert isinstance(obj.__root__, JSONWebEncryptionFlattened)
    assert len(obj.recipients) == 1


def test_protected_and_unprotected_header_must_be_disjoint(token: jwcrypto.jwe.JWE):
    obj = json.loads(token.serialize())
    JSONWebEncryption.parse_obj(obj)
    obj.update({
        'protected': b64encode(str.encode(json.dumps({'alg': 'foo'}))),
        'unprotected': {'alg': 'foo'},
    })
    with pytest.raises(pydantic.ValidationError):
        JSONWebEncryption.parse_obj(obj)


def test_protected_and_recipient_header_must_be_disjoint(token: jwcrypto.jwe.JWE):
    obj = json.loads(token.serialize())
    JSONWebEncryption.parse_obj(obj)
    obj.update({
        'protected': b64encode(str.encode(json.dumps({'foo': 1}))),
    })
    with pytest.raises(pydantic.ValidationError):
        JSONWebEncryption.parse_obj(obj)


def test_unprotected_and_recipient_header_must_be_disjoint(token: jwcrypto.jwe.JWE):
    obj = json.loads(token.serialize())
    JSONWebEncryption.parse_obj(obj)
    obj.update({
        'unprotected': {'bar': 2}
    })
    with pytest.raises(pydantic.ValidationError):
        JSONWebEncryption.parse_obj(obj)


def test_header_attribute_contains_merged(token: jwcrypto.jwe.JWE):
    obj = json.loads(token.serialize())
    obj.update({
        'unprotected': {'baz': 2}
    })
    jwe = JSONWebEncryption.parse_obj(obj)
    assert 'baz' in jwe.header


def test_flattened_protected_and_unprotected_header_must_be_disjoint(token_single: jwcrypto.jwe.JWE):
    obj = json.loads(token_single.serialize())
    JSONWebEncryption.parse_obj(obj)
    obj.update({
        'protected': b64encode(str.encode(json.dumps({'alg': 'foo'}))),
        'unprotected': {'alg': 'foo'},
    })
    with pytest.raises(pydantic.ValidationError):
        JSONWebEncryption.parse_obj(obj)


def test_flattened_protected_and_recipient_header_must_be_disjoint(token_single: jwcrypto.jwe.JWE):
    obj = json.loads(token_single.serialize())
    JSONWebEncryption.parse_obj(obj)
    obj.update({
        'header': {'bar': 1},
    })
    with pytest.raises(pydantic.ValidationError):
        JSONWebEncryption.parse_obj(obj)


def test_flattened_unprotected_and_recipient_header_must_be_disjoint(token_single: jwcrypto.jwe.JWE):
    obj = json.loads(token_single.serialize())
    JSONWebEncryption.parse_obj(obj)
    obj.update({
        'header': {'bar': 2}
    })
    with pytest.raises(pydantic.ValidationError):
        JSONWebEncryption.parse_obj(obj)


def test_parse_compact():
    payload = b'Hello world!'
    jwk = jwcrypto.jwk.JWK.generate(kty='oct', size=256)
    jwe = jwcrypto.jwe.JWE(
        payload,
        json_encode({
            'alg': "A256KW",
            'enc': "A256CBC-HS512"
        })
    )
    jwe.add_recipient(jwk)
    obj = JSONWebEncryption.parse_compact(
        str.encode(jwe.serialize(compact=True))
    )


# Copyright 2022 Cochise Ruhulessin
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Declares :class:`IProvider`."""
import importlib
from typing import cast
from typing import Any
from typing import Awaitable
from typing import Callable
from typing import Generator
from typing import TypeAlias
from typing import Union

import dsnparse
from unimatrix.exceptions import CanonicalException

from .generatekeyoperation import GenerateKeyOperation
from .ikeyinspector import IKeyInspector
from .undecryptable import Undecryptable # type: ignore


class IProvider:
    __module__: str = 'cbra.types'
    SignResult: TypeAlias = bytes | Awaitable[bytes]
    Undecryptable: type[CanonicalException] = Undecryptable
    _default_providers: dict[str, str] = {
        'azure': 'ckms.ext.azure.AzureProvider',
        'google': 'ckms.ext.google.GoogleProvider',
        'local': 'ckms.core.local.LocalProvider',
        'random': 'ckms.core.random.RandomProvider',
        'ssh-agent': 'ckms.ext.ssh.provider.SSHAgentProvider',
    }
    _providers: dict[str, 'IProvider'] = {}
    inspector: IKeyInspector

    @classmethod
    def __get_validators__(cls) -> Generator[Callable[..., 'IProvider'], None, None]:
        yield cls.validate

    @classmethod
    def validate(cls, value: Union[str, 'IProvider']):
        if isinstance(value, str):
            value = cls.get(value)
        return value

    @staticmethod
    def get(name: str) -> 'IProvider':
        """Return the provider implementation specified by the `name`
        argument.
        """
        if name not in IProvider._providers:
            ProviderImpl = cast(type[IProvider], IProvider.load_impl(name))
            IProvider._providers[name] = ProviderImpl()
        return IProvider._providers[name]

    @staticmethod
    def load_impl(name: str) -> type['IProvider']:
        """Loads the implementation class `name`."""
        module_name, symbol = str.rsplit(IProvider._default_providers[name], '.', 1)
        return getattr(importlib.import_module(module_name), symbol)

    @staticmethod
    def register(name: str, qualname: str) -> None:
        """Registers a new provider implementation under the given name."""
        IProvider._default_providers[name] = qualname # pragma: no cover

    def calculate_kid(self, value: Any) -> Any:
        return self.inspector.calculate_kid(value)

    def jwk(self, spec: Any, private: bool, **claims: Any) -> Any:
        """Create a JWK from a key specification."""
        raise NotImplementedError

    def parse_dsn(self, dsn: dsnparse.ParseResult) -> Any:
        """Parse a key specification from a Data Source Name (DSN)."""
        raise NotImplementedError

    def decrypt(
        self,
        key: Any,
        algorithm: str,
        ciphertext: Any
    ) -> Any | Awaitable[Any]:
        """Decrypt a cipher test using the given algorithm."""
        raise NotImplementedError

    def encrypt(
        self,
        key: Any,
        algorithm: str,
        plaintext: Any,
        **kwargs: Any
    ) -> Any | Awaitable[Any]:
        """Encrypt message using the given algorithm."""
        raise NotImplementedError

    def wrap(
        self,
        key: Any,
        algorithm: str,
        wrap: Any,
        **kwargs: Any
    ) -> Any | Awaitable[Any]:
        """Wrap a key using the given algorithm."""
        raise NotImplementedError

    async def fetch(self, blob: Any) -> bytes:
        """Fetches the content of the remote blob using the parameters
        specified. Return a byte-string holding the content.
        """
        raise NotImplementedError

    async def generate(self, op: GenerateKeyOperation) -> Any:
        """Generate a new private key."""
        raise NotImplementedError

    async def load(self, key: Any) -> None:
        """Loads the key metadata and updates its properties."""
        raise NotImplementedError

    async def random(self, length: int) -> bytes:
        """Return a random byte-sequence of the given length."""
        raise NotImplementedError

    def get_hashing_algorithm(self, algorithm: str) -> str | None:
        """Return a string indicating the hashing algorithm."""
        raise NotImplementedError

    def get_key_identifier(self, spec: Any) -> str: # type: ignore
        """Generate a key identifier for the given private key."""
        raise NotImplementedError

    def get_public_key(self, spec: Any) -> Any | None:
        """Return a :class:`~ckms.core.types.PublicKeySpecification`
        instance representing the public part of an asymmetric keypair,
        or ``None`` if `spec` is not an asymmetric key.
        """
        raise NotImplementedError

    def parse_spec(self, spec: dict[str, Any]) -> Any:
        """Parse a key specification from a dictionary."""
        raise NotImplementedError

    def sign(
        self,
        key: Any,
        algorithm: str,
        message: Any
    ) -> bytes | Awaitable[bytes]:
        """Sign message using the given algorithm."""
        raise NotImplementedError

    def validate_provider(self) -> Any: # type: ignore
        """Validate the parameters as required by the provider and return
        the updated instance.
        """
        raise NotImplementedError

    def verify(
        self,
        key: Any,
        algorithm: str,
        message: Any,
        signature: bytes
    ) -> bool | Awaitable[bool]:
        """Verify a digital signature, specified by the operation."""
        raise NotImplementedError


# pylint: skip-file
import time
from typing import Any

import pytest

from ckms.utils import b64encode
from ckms.types import ClaimSet
from ckms.types import ClaimTypeError
from ckms.types import JSONWebToken
from ckms.types import MalformedPayload
from ckms.types import MaximumAgeExceeded
from ckms.types import MissingRequiredClaim
from ckms.types import TokenExpired
from ckms.types import TokenNotEffective
from ckms.types import WrongAudience


@pytest.mark.parametrize("claim", [
    "jti",
    "iss",
    "aud",
    "iat",
    "nbf",
    "exp"
])
def test_strict(claim: str):
    claims = JSONWebToken.strict(
        iss="foo",
        aud="foo",
        ttl=600
    )
    assert getattr(claims, claim, None) is not None


def test_strict_sets_standardized_claim():
    claims = JSONWebToken.strict(
        iss="foo",
        aud="foo",
        ttl=600,
        sub="foo"
    )
    assert claims.sub == "foo" # type: ignore


def test_strict_sets_private_claim():
    claims = JSONWebToken.strict(
        iss="foo",
        aud="foo",
        ttl=600,
        foo="foo"
    )
    assert claims.extra['foo'] == "foo"


def test_dict_only_returns_present_claims():
    claims = JSONWebToken(foo=1) # type: ignore
    data = claims.dict()
    assert len(data) == 1
    assert 'foo' in data


def test_json_only_returns_present_claims():
    claims = JSONWebToken(foo=1) # type: ignore
    data = claims.json()
    assert data == '{"foo": 1}'


def test_get_attribute():
    claims = JSONWebToken(foo=1) # type: ignore
    assert claims.extra['foo'] == 1


def test_strict_sets_ttl():
    claims = JSONWebToken.strict(
        iss="foo",
        aud="foo",
        ttl=600
    )
    assert claims.expires_in() == 600 # type: ignore
    time.sleep(1)
    assert claims.expires_in() == 599 # type: ignore


def test_serialize():
    claims = JSONWebToken.frompayload(bytes(JSONWebToken(iss='foo')))
    assert claims.iss == 'foo' # type: ignore


def test_frompayload():
    claims = JSONWebToken.frompayload(b64encode(b'{"foo": 1}'))
    assert claims.extra.get('foo') == 1


def test_frompayload_raised_malformed():
    with pytest.raises(MalformedPayload):
        JSONWebToken.frompayload(b'f')


@pytest.mark.parametrize("Model", [JSONWebToken])
@pytest.mark.parametrize("claim,value", [
    ('exp', 'a'),
    ('iat', 'a'),
    ('nbf', 'a'),
])
def test_invalid_type_raises_invalidclaimtype(
    Model: type[ClaimSet],
    claim: str,
    value: Any
):
    with pytest.raises(ClaimTypeError):
        Model.parse_obj({claim: value})


def test_verify_raises_missing_required_claim():
    claims = JSONWebToken(foo=1) # type: ignore
    with pytest.raises(MissingRequiredClaim):
        claims.verify(required={'bar'})
    claims.verify(required={'foo'})


def test_validate_exp_raises():
    claims = JSONWebToken(exp=1)
    with pytest.raises(TokenExpired):
        claims.verify(now=2)


def test_exp_from_expires_in():
    claims = JSONWebToken()
    assert claims.expires_in(ttl=5, now=0) == 5
    assert claims.exp == 5


def test_validate_nbf_raises():
    claims = JSONWebToken(nbf=2)
    with pytest.raises(TokenNotEffective):
        claims.verify(now=1)


def test_aud_is_set_when_not_provided():
    claims = JSONWebToken()
    assert isinstance(claims.aud, set)


def test_aud_is_set_when_string():
    claims = JSONWebToken(aud="foo") # type: ignore
    assert isinstance(claims.aud, set)
    assert claims.aud == {"foo"}


def test_aud_is_set_when_list():
    claims = JSONWebToken(aud=["foo", "bar"]) # type: ignore
    assert isinstance(claims.aud, set)
    assert claims.aud == {"foo", "bar"}


def test_aud_not_list_or_str_raises():
    with pytest.raises(ClaimTypeError):
        JSONWebToken(aud=1) # type: ignore


def test_aud_must_be_all_strings():
    with pytest.raises(ClaimTypeError):
        JSONWebToken(aud=[1, "foo"]) # type: ignore


def test_aud_must_intersect():
    claims = JSONWebToken(aud=["foo"]) # type: ignore
    with pytest.raises(WrongAudience):
        claims.verify(audience={"bar"})
    claims.verify(audience={"foo", "bar"})


def test_verify_strict():
    with pytest.raises(MissingRequiredClaim):
        claims = JSONWebToken(iat=5, exp=10)
        claims.verify(strict=True, now=9)


def test_verify_max_age_exceeded():
    claims = JSONWebToken(iat=1)
    with pytest.raises(MaximumAgeExceeded):
        claims.verify(max_age=1, now=3)


def test_verify_max_age():
    claims = JSONWebToken(iat=1)
    claims.verify(max_age=2, now=2)


def test_selfsigned():
    claims = JSONWebToken(sub='foo', iss='foo')
    assert claims.is_selfsigned()


# Copyright 2022 Cochise Ruhulessin
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from typing import Any

import pydantic
from cryptography.hazmat.primitives.asymmetric.utils import encode_dss_signature
from pyasn1.type import univ
from pyasn1_modules import rfc5280

from ckms.ext import asn1
from ckms.ext.rfc5480.types import ECParameters
from ckms.ext.asn1.fields import ObjectIdentifier
from ckms.utils import normalize_ec_signature
from ckms.utils import bytes_to_number


class AlgorithmIdentifier(asn1.ASN1Model):
    __asn1spec__: type[rfc5280.AlgorithmIdentifier] = rfc5280.AlgorithmIdentifier
    name: ObjectIdentifier = pydantic.Field(..., alias='algorithm')
    parameters: univ.Any | ECParameters | None

    @pydantic.root_validator(pre=True)
    def preprocess(cls, values: dict[str, Any]) -> dict[str, Any]:
        params = values.get('parameters')
        if params is None:
            return values
        if not isinstance(params, ECParameters) and params.isValue:
            params = asn1.decode_der(value=params)
            if params:
                # TODO: This assumes that there are params only with EC
                values['parameters'] = ECParameters.parse_obj({
                    'namedCurve': str(params)
                })
        return values

    @pydantic.validator('name', pre=True)
    def validate_name(cls, value: str | univ.ObjectIdentifier) -> str:
        return str(value)

    def encode_signature(self, signature: bytes) -> bytes:
        if isinstance(self.parameters, ECParameters):
            assert len(signature) in {64, 96, 132}, len(signature)
            l = int(len(signature) / 2)
            signature = encode_dss_signature(
                r=bytes_to_number(signature[:l]),
                s=bytes_to_number(signature[l:])
            )
        return signature

    def normalize_signature(self, signature: bytes) -> bytes:
        if isinstance(self.parameters, ECParameters):
            curve = self.parameters.curve.get()
            signature = normalize_ec_signature(
                l=(curve.key_size + 7) // 8,
                sig=signature
            )
        return signature


# Copyright 2022 Cochise Ruhulessin
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from typing import Any
from typing import Generator


class IKeySpecification:
    __module__: str = 'ckms.types'
    algorithm: Any
    provider: Any

    def can_decrypt(self) -> bool:
        """Return a boolean indicating if the key may be used to decrypt."""
        raise NotImplementedError

    def can_encrypt(self) -> bool:
        """Return a boolean indicating if the key may be used to encrypt."""
        raise NotImplementedError

    def can_wrap(self) -> bool:
        """Return a boolean indicating if the key can wrap another key."""
        raise NotImplementedError

    def is_loaded(self) -> bool:
        """Return a boolean indicating if the specification is loaded into
        memory. Note that this regards the *metadata* and not necessarily
        the actual key material.
        """
        raise NotImplementedError

    def get_private_bytes(self) -> bytes:
        """Return the private bytes that may be used for decryption or
        signing.
        """
        raise NotImplementedError

    def get_private_key(self) -> Any:
        """Return the private key that may be used for decryption or
        signing.
        """
        raise NotImplementedError

    def get_public_key(self) -> Any:
        """Return the public key that may be used for encryption or
        verifcation.
        """
        raise NotImplementedError

    def __await__(self) -> Generator[Any, None, 'IKeySpecification']:
        return self.load().__await__()

    async def load(self) -> 'IKeySpecification':
        """Inspect the configuration and load the metadata regarding the
        key.
        """
        raise NotImplementedError


# Copyright 2018 Cochise Ruhulessin
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Declares :class:`JOSEObject`."""
import binascii
import json
from typing import Any
from typing_extensions import Annotated

import pydantic

from ckms.types import MalformedObject
from ckms.types import MalformedPayload
from ckms.utils import b64decode
from .jsonwebencryption_ import JSONWebEncryption
from .jsonwebsignature import JSONWebSignature


class JOSEObject(pydantic.BaseModel):
    __root__: Annotated[JSONWebEncryption | JSONWebSignature, pydantic.Field(discriminator='kind')]

    @classmethod
    def deserialize_compact(cls, protected: str, *segments: str) -> dict[str, Any]:
        """Deserialize a JWE or JWS that uses the compact serialization format."""
        obj: dict[str, Any]
        assert len(segments) in {2,4}
        if len(segments) == 2:
            payload, signature = segments
            obj = {
                'kind': 'JWS',
                'payload': payload,
                'signatures': [
                    {
                        'protected': protected,
                        'signature': signature
                    }
                ]
            }
        elif len(segments) == 4:
            encryption_key, iv, ciphertext, tag = segments
            obj = {
                'kind': 'JWE',
                'protected': protected,
                'recipients': [{
                    'encrypted_key': encryption_key,
                }],
                'ciphertext': ciphertext,
                'tag': tag,
                'compact': True
            }
            if iv:
                obj['iv'] = iv
        else: # pragma: no cover
            raise ValueError("Invalid number of segments in token.")
        return obj

    @classmethod
    def deserialize(cls, token: bytes | str) -> dict[str, Any]:
        """Parse a serialized JOSE object (JWE, JWS or nested JWT) from a
        byte-sequence.
        """
        if isinstance(token, bytes):
            token = bytes.decode(token, 'ascii')
        is_compact = token.find('.') != -1
        if is_compact:
            obj = {}
            if token.count('.') not in {2, 4}:
                raise MalformedObject(
                    detail="Invalid number of segments in token.",
                    hint="JOSE Compact Serialization has 3 segments for JWS, 5 for JWE."
                )
            protected, *segments = str.split(token, '.')
            obj = cls.deserialize_compact(protected, *segments)
        else:
            try:
                if not (token[0] == '{' and token[-1] == '}'):
                    # Assume here that if the token does not start and end with a bracket,
                    # we are dealing with a Base64-encoded string or byte-sequence.
                    token = bytes.decode(b64decode(token), 'utf-8')
            except (binascii.Error, IndexError, ValueError):
                raise MalformedObject

            try:
                obj: dict[str, Any] = json.loads(token)
            except ValueError:
                raise MalformedPayload

            # Determine if we are parsing a flattened JWE/JWS and convert
            # it to a general JSON object.
            if 'signature' in obj: # Flattened JWS
                obj['signatures'] = [
                    {
                        'header': obj.pop('header', {}),
                        'protected': obj.pop('protected', None),
                        'signature': obj.pop('signature', None),
                    }
                ]
                assert set(obj.keys()) >= {'payload', 'signatures'}, obj

            if 'encrypted_key' in obj: # Flattened JWE
                obj['recipients'] = [
                    {
                        'header': obj.pop('header', {}),
                        'encrypted_key': obj.pop('encrypted_key', None)
                    }
                ]

        if obj.get('recipients'):
            obj['kind'] = 'JWE'
        if obj.get('signatures'):
            obj['kind'] = 'JWS'
        obj['token'] = token
        return obj

    @classmethod
    def parse(cls, token: bytes | str) -> JSONWebEncryption | JSONWebSignature:
        obj = cls.deserialize(token=token)
        return cls.parse_obj(obj).__root__


# Copyright 2022 Cochise Ruhulessin
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import base64
from typing import Any

import httpx
import pydantic
from cryptography import x509
from cryptography.hazmat.primitives.serialization import Encoding

from ckms.utils import b64encode_sha1
from ckms.utils import b64encode_sha256
from ckms.utils import certificate_chain


class Certificate(pydantic.BaseModel):
    uri: str
    _certificates: list[tuple[x509.Certificate, bytes]] = pydantic.PrivateAttr(default=[])

    @property
    def x5c(self) -> list[str]:
        return [
            bytes.decode(base64.b64encode(der))
            for _, der in self._certificates
        ]

    @property
    def x5t(self) -> str:
        return bytes.decode(b64encode_sha1(self._certificates[0][1]))

    @property
    def x5t_sha256(self) -> str:
        return bytes.decode(b64encode_sha256(self._certificates[0][1]))

    @property
    def x5u(self) -> str:
        return self.uri

    def claims(self) -> dict[str, Any]:
        return {
            'x5u': self.uri,
            'x5t': self.x5t,
            'x5t#S256': self.x5t_sha256,
            'x5c': self.x5c
        }

    async def load(self) -> None:
        """Fetch the certificate from :attr:`uri`. The content is
        expected to be a PEM-encoded X.509 certificates. Additional
        certificates are the chain.
        """
        async with httpx.AsyncClient() as client:
            response = await client.get(url=self.uri) # type: ignore
            for pem in certificate_chain(response.content):
                crt = x509.load_pem_x509_certificate(pem)
                self._certificates.append(
                    (crt, crt.public_bytes(Encoding.DER))
                )

    class Config:
        arbitrary_types_allowed: bool = True


# Copyright 2022 Cochise Ruhulessin
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import pytest

from ckms.jose import PayloadCodec
from ckms.core import Keychain
from ckms.core.models import KeySpecification
from ckms.types import InvalidSignature
from .conftest import SIGNING_KEYS_ASYMMETRIC
from .conftest import SIGNING_KEYS_SYMMETRIC



@pytest.mark.parametrize("spec", [
    *SIGNING_KEYS_ASYMMETRIC,
    *SIGNING_KEYS_SYMMETRIC
])
class TestProviderJWS:

    @pytest.mark.asyncio
    async def test_sign_and_verify_compact(
        self,
        signer: Keychain,
        verifier: Keychain,
        spec: KeySpecification
    ):
        assert spec.kid is not None
        c1 = PayloadCodec(
            signer=signer,
            verifier=verifier,
            signing_keys=[spec.kid]
        )
        c2 = PayloadCodec()
        jws = await c1.encode({'foo': 1})
        assert await c1.decode(jws, verify=True)
        with pytest.raises(InvalidSignature):
            await c2.decode(jws, verify=True)


# Copyright 2018 Cochise Ruhulessin
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# type: ignore
from typing import Any

import pydantic
from pyasn1.type import univ
from pyasn1_modules import rfc2315

from ckms.ext.x509.types import AlgorithmIdentifier
from ckms.ext.x509.types import Certificate
from .contentinfo import ContentInfo
from .signerinfo import SignerInfo


class SignedData(pydantic.BaseModel):
    version: int
    digest_algorithms: set[AlgorithmIdentifier]
    content_info: ContentInfo
    certificates: set[Certificate] | None = None
    signers: list[SignerInfo] = pydantic.Field(..., alias='signerInfos')

    @property
    def content(self) -> bytes:
        return self.content_info.content

    @pydantic.root_validator(pre=True)
    def preprocess(cls, values: dict[str, Any]) -> dict[str, Any]:
        if isinstance(values.get('version'), rfc2315.Version):
            values['version'] = int(values['version'])
        algorithms = values['digest_algorithms'] = list(values.get('digestAlgorithms') or [])
        for i, algorithm in enumerate(algorithms):
            algorithms[i] = AlgorithmIdentifier(str(algorithm['algorithm']))
        values['content_info'] = values.get('contentInfo')
        return values

    @pydantic.validator('certificates', pre=True)
    def validate_certificates(
        cls,
        value: set[Certificate] | rfc2315.ExtendedCertificatesAndCertificates | None
    ) -> list[Any] | set[Certificate] | None:
        if isinstance(value, univ.SetOf):
            value = [x for x in value]
        return value

    @pydantic.validator('signers', pre=True)
    def validate_signers(
        cls,
        value: set[SignerInfo] | rfc2315.SignerInfo | None
    ) -> list[Any] | set[SignerInfo] | None:
        if isinstance(value, univ.SetOf):
            value = [x for x in value]
        return value


# Copyright 2022 Cochise Ruhulessin
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Declares :class:`LocalProvider`."""
import functools
import hmac
import inspect
from typing import cast
from typing import Any

import aiofiles
import dsnparse
from ckms.types import CipherText
from ckms.types import HMACOperation
from cryptography.exceptions import InvalidSignature
from cryptography.exceptions import InvalidTag
from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.primitives.asymmetric import ec
from cryptography.hazmat.primitives.asymmetric import padding
from cryptography.hazmat.primitives.asymmetric import rsa
from cryptography.hazmat.primitives.asymmetric import utils
from cryptography.hazmat.primitives.ciphers import algorithms
from cryptography.hazmat.primitives.ciphers import AEADCipherContext
from cryptography.hazmat.primitives.keywrap import aes_key_unwrap

import ckms.types
from ckms import core
from ckms.core.provider import Provider
from ckms.utils import bytes_to_number
from ckms.utils import number_to_bytes
from .types import HMAC
from .types import Key
from .models import KeySpecification


class LocalProvider(Provider):
    """A :class:`~ckms.core.Provider` implementation that uses the
    local CPU to perform cryptographic operations.
    """
    require_algorithm: bool = True
    supported_rsa_algorithms: set[str] = {
        'RS256',
        'RS384',
        'RS512',
        'PS256',
        'PS384',
        'PS512',
        'RSA-OAEP',
        'RSA-OAEP-256',
        'RSA-OAEP-384',
        'RSA-OAEP-512',
    }
    supported_ed_curves: set[str] = {"Ed448", "X448", "Ed25519", "X25519"}
    spec: type[KeySpecification] = KeySpecification # type: ignore
   
    async def generate(
        self,
        op: ckms.types.GenerateKeyOperation
    ) -> Any: # type: ignore
        """Select the proper class and generate a new key."""
        result = self._generate(op.spec)
        if inspect.isawaitable(result):
            result = await result
        return result


    def parse_dsn(self, dsn: dsnparse.ParseResult) -> KeySpecification:
        return {
            **dsn.query, # type: ignore
            'provider': 'local',
            'key': {'path': dsn.path[1:]} # type: ignore
        }

    @functools.singledispatchmethod
    def _generate(self, spec: Any) -> Any:
        raise NotImplementedError(repr(spec))

    @_generate.register
    async def _generate_aes(self, spec: ckms.types.GenerateAES) -> Any:
        return algorithms.AES(await self.random(spec.length))

    @_generate.register
    def _generate_ec(self, spec: ckms.types.GenerateEllipticCurve) -> Any:
        generate, curve = self.inspector.get_ec_class(spec.curve)
        return generate(curve)

    @_generate.register
    async def _generate_hmac(self, spec: ckms.types.GenerateHMAC) -> Any:
        return HMAC(await self.random(spec.length))

    @_generate.register
    def _generate_rsa(self, spec: ckms.types.GenerateRSA) -> Any:
        return rsa.generate_private_key(
            public_exponent=65537,
            key_size=spec.length
        )

    @_generate.register
    def _generate_okp(self, spec: ckms.types.GenerateOKP) -> Any:
        cls = self.inspector.get_ed_class(spec.curve)
        return cls.generate()

    def get_digest_algorithm(self, algorithm: str) -> hashes.HashAlgorithm:
        """Return an instance of the selected `algorithm`."""
        assert algorithm in {'sha1', 'sha256', 'sha384', 'sha512'}, algorithm
        return getattr(hashes, str.upper(algorithm))()

    def get_key_identifier(self, spec: models.KeySpecification) -> str: # type: ignore
        """Generate a key identifier for the given private key."""
        return self.inspector.calculate_kid(spec.key) # type: ignore

    def get_padding(self, scheme: str, digest: str) -> padding.AsymmetricPadding:
        """Return an instance of the specified padding."""
        algorithm = self.get_digest_algorithm(digest)
        if scheme in {'EMSA-PSS', 'PSS'}:
            instance = padding.PSS(
                mgf=padding.MGF1(algorithm),
                salt_length=algorithm.digest_size
            )
        elif scheme in {'EMSA-PKCS1-v1_5', 'PKCS1v15'}:
            instance = padding.PKCS1v15()
        elif scheme in {'OAEP'}:
            instance = padding.OAEP(
                mgf=padding.MGF1(algorithm),
                algorithm=algorithm,
                label=None
            )
        else: # pragma: no cover
            raise ValueError(f"Unsupported scheme: {scheme}")
        return instance

    def get_public_key(self, spec: KeySpecification) -> Any | None:
        return spec.get_public_key()

    def unwrap_aes_wrap(self, op: ckms.types.AESKeywrapOperation) -> bytes:
        spec = op.get_keyspec()
        return aes_key_unwrap(
            wrapping_key=spec.get_private_bytes(),
            wrapped_key=bytes(op.content)
        )

    async def decrypt_aes(self, op: ckms.types.AESOperation) -> bytes:
        ct = op.get_ciphertext()
        if not op.get_initialization_vector():
            raise self.Undecryptable
        cipher = op.get_cipher()
        d = cipher.decryptor()
        if isinstance(d, AEADCipherContext) and ct.aad is not None:
            d.authenticate_additional_data(ct.aad)
        try:
            pt = d.update(await op.get_message()) + d.finalize()
        except InvalidTag:
            raise self.Undecryptable(
                detail=(
                    "Unable to decrypt the ciphertext using  Advanced "
                    "Encryption Standard (AES)."
                )
            )
        if op.padding:
            raise NotImplementedError
        return pt

    async def encrypt_aes(self, op: ckms.types.AESOperation) -> CipherText:
        iv = await self.random(12)
        cipher = op.get_cipher(iv=iv)
        e = cipher.encryptor()
        data = await op.get_message()
        if isinstance(e, AEADCipherContext) and op.aad is not None:
            e.authenticate_additional_data(op.aad)
        if op.padding:
            raise NotImplementedError
        buf = e.update(data) + e.finalize()
        return CipherText(
            aad=op.aad,
            buf=buf,
            iv=iv,
            tag=getattr(e, 'tag', None)
        )

    async def unwrap(self, op: ckms.types.KeywrapOperation) -> bytes:
        assert op.mode == 'aes'
        algorithm = op.get_private_key()
        return aes_key_unwrap(
            wrapping_key=algorithm.key,
            wrapped_key=await op.get_message()
        )

    async def decrypt_rsa(self, op: ckms.types.RSAOperation) -> bytes:
        key = op.get_private_key()
        return key.decrypt(
            ciphertext=await op.get_message(),
            padding=self.get_padding(op.padding, op.digest)
        )

    async def encrypt_rsa(self, op: ckms.types.RSAOperation) -> ckms.types.CipherText:
        key = op.get_public_key()
        return ckms.types.CipherText(
            buf=key.encrypt(
                await op.get_message(),
                padding=self.get_padding(op.padding, op.digest)
            )
        )

    async def fetch(self, blob: core.RemoteBlob) -> bytes:
        """Fetches the content of the remote blob using the parameters
        specified. Return a byte-string holding the content.
        """
        blob = Key.parse_obj(blob.dict()) # type: ignore
        async with aiofiles.open(blob.path, 'rb') as f: # type: ignore
            return await f.read() # type: ignore

    async def sign_ec(self, op: ckms.types.EllipticCurveOperation) -> bytes:
        key = cast(ec.EllipticCurvePrivateKey, op.spec.key)
        n = (key.curve.key_size + 7) // 8
        sig = key.sign(
            data=await op.get_digest(),
            signature_algorithm=ec.ECDSA(
                utils.Prehashed(self.get_digest_algorithm(op.digest))
            )
        )
        r, s = utils.decode_dss_signature(sig)
        return number_to_bytes(r, n) + number_to_bytes(s, n)

    async def sign_eddsa(self, op: ckms.types.EdDSAOperation) -> bytes:
        return op.spec.key.sign(await op.get_message())

    async def sign_hmac(self, op: ckms.types.HMACOperation) -> bytes:
        assert op.digestmod is not None
        mac = hmac.new(
            key=op.spec.key,
            digestmod=op.digestmod
        )
        async for chunk in op.chunks():
            mac.update(chunk)
        return mac.digest()

    async def sign_rsa(self, op: ckms.types.RSAOperation) -> bytes:
        return op.spec.key.sign(
            await op.get_digest(),
            padding=self.get_padding(op.padding, op.digest),
            algorithm=utils.Prehashed(self.get_digest_algorithm(op.digest))
        )

    async def verify_ec(self, op: ckms.types.EllipticCurveOperation) -> bool:
        if op.signature is None:
            raise TypeError("Signature can not be None.")
        public = op.get_public_key()
        n = (public.curve.key_size + 7) // 8
        try:
            public.verify(
                signature=utils.encode_dss_signature(
                    bytes_to_number(op.signature[:n]),
                    bytes_to_number(op.signature[n:]),
                ),
                data=await op.get_digest(),
                signature_algorithm=ec.ECDSA(
                    utils.Prehashed(self.get_digest_algorithm(op.digest))
                )
            )
            return True
        except InvalidSignature:
            return False

    async def verify_eddsa(self, op: ckms.types.EdDSAOperation) -> bool:
        key = op.spec.get_public_key()
        try:
            key.verify(op.signature, await op.get_message())
            return True
        except InvalidSignature:
            return False

    async def verify_hmac(self, op: HMACOperation) -> bool:
        assert op.digestmod is not None
        assert op.signature is not None
        mac = hmac.new(
            key=op.spec.key,
            digestmod=op.digestmod
        )
        async for chunk in op.chunks():
            mac.update(chunk)
        return hmac.compare_digest(op.signature, mac.digest())

    async def verify_rsa(self, op: ckms.types.RSAOperation) -> bool:
        public = op.get_public_key()
        try:
            public.verify(
                signature=op.signature,
                data=await op.get_digest(),
                padding=self.get_padding(op.padding, op.digest),
                algorithm=utils.Prehashed(self.get_digest_algorithm(op.digest))
            )
            return True
        except InvalidSignature:
            return False


# Copyright 2022 Cochise Ruhulessin
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import pydantic
from pyasn1.type import univ
from pyasn1_modules import rfc5280


class BaseCertificateExtension(pydantic.BaseModel):
    ext_id: str
    critical: bool = True

    def asn1(self) -> rfc5280.Extension:
        obj = rfc5280.Extension()
        obj['extnID'] = self.ext_id
        obj['critical'] = self.critical
        obj['extnValue'] = self.get_encoded_extension()
        return obj

    def get_encoded_extension(self) -> univ.OctetString:
        raise NotImplementedError


# Copyright 2018 Cochise Ruhulessin
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# type: ignore
import inspect
import secrets
from typing import cast
from typing import Any

from cryptography import x509
from pyasn1.codec.der import decoder
from pyasn1.codec.der import encoder
from pyasn1_modules import rfc3161

from ckms.types import Digest
from ckms.types import Message
from ckms.types import IHTTPClient
from .types import AlgorithmIdentifier
from .types import Timestamp
from .types import TimeStampResp


class TimestampingAuthority:
    """Provides an interface to create a timestamp using a specific authority."""
    __module__: str = 'ckms.ext.tsa'
    certificate: x509.Certificate | None = None
    digest: str
    url: str

    def __init__(
        self,
        *,
        url: str,
        crtpath: str | None = None,
        digest: str = 'sha1'
    ):
        if crtpath:
            self.certificate = x509.load_pem_x509_certificate(open(crtpath, 'rb').read())
        self.digest = digest
        self.url = url

    def create_request(self, data: bytes, digest_name: str) -> rfc3161.TimeStampReq:
        """Return a byte-sequence containing a DER-encoded :rfc:`3161`
        ``TimeStampReq`` object.
        """
        message = rfc3161.MessageImprint()
        message.setComponentByPosition(0, AlgorithmIdentifier.new(digest_name))
        message.setComponentByPosition(1, data)
        request = rfc3161.TimeStampReq()
        request.setComponentByPosition(0, 'v1')
        request.setComponentByPosition(1, message)
        request.setComponentByPosition(3, self.nonce())
        request.setComponentByPosition(4)
        #with open('request-data.tsq', 'wb') as f: f.write(data)
        return request

    def nonce(self) -> int:
        return int.from_bytes(secrets.token_bytes(64), 'big')

    async def timestamp(
        self,
        *,
        client: Any,
        data: bytes | Digest | Message,
        digest: str | None = None
    ) -> TimeStampResp:
        """Create a timestamp of the given input data."""
        client = cast(IHTTPClient, client)
        if not isinstance(data, (Digest, Message)):
            data = Message(buf=data, digest=digest or self.digest)
        assert isinstance(data, Message)
        request = self.create_request(
            data=await data.digest(),
            digest_name=data.digestmod
        )
        response = client.post(
            url=self.url,
            content=encoder.encode(request),
            headers={
                'Content-Type': 'application/timestamp-query'
            }
        )
        if inspect.isawaitable(response):
            response = await response
        obj, _ = decoder.decode(response.content, asn1Spec=rfc3161.TimeStampResp())
        if _:
            raise ValueError("Invalid response from TSA.")
        #with open('response.tsq', 'wb') as f: f.write(response.content)
        return TimeStampResp.parse_obj({
            **obj,
            'content': response.content,
            'request': request
        })


# Copyright 2018 Cochise Ruhulessin
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from typing import Awaitable
from typing import TypeAlias

from .aesalgorithmtype import AESAlgorithmType
from .aeskeywrapoperation import AESKeywrapOperation
from .aesoperation import AESOperation
from .algorithm import Algorithm
from .asymmetricsigningoperation import AsymmetricSigningOperation
from .audiencetype import AudienceType
from .ciphertext import CipherText
from .claimset import ClaimSet
from .cryptographicfailure import *
from .cryptographypublickeytype import CryptographyPublicKeyType
from .data import Data
from .decrypter import Decrypter
from .digest import Digest
from .edsaoperation import EdDSAOperation
from .edwardscurvealgorithmtype import EdwardsCurveAlgorithmType
from .edwardscurvetype import EdwardsCurveType
from .ellipticcurvealgorithmtype import EllipticCurveAlgorithmType
from .ellipticcurvetype import EllipticCurveType
from .ellipticcurveoperation import EllipticCurveOperation
from .encrypter import Encrypter
from .encryptoperation import EncryptOperation
from .hmacalgorithmtype import HMACAlgorithmType
from .hmacoperation import HMACOperation
from .icontentencryptionkey import IContentEncryptionKey
from .forbiddenoperation import ForbiddenOperation
from .generatekeyspec import *
from .generatekeyoperation import GenerateKeyOperation
from .ihasher import IHasher
from .ihttpclient import IHTTPClient
from .ikeychain import IKeychain
from .ikeyinspector import IKeyInspector
from .ikeyspecification import IKeySpecification
from .invalidcredential import *
from .invalidtoken import *
from .ioperationperformer import IOperationPerformer
from .iprovider import IProvider
from .jsonwebkey import JSONWebKey
from .jsonwebkeyset import JSONWebKeySet
from .jsonwebtoken import JSONWebToken
from .keyalgorithmtype import KeyAlgorithmType
from .keyoperationtype import KeyOperationType
from .keywrapoperation import KeywrapOperation
from .keyusetype import KeyUseType
from .malformed import *
from .message import Message
from .operation import Operation
from .plaintext import PlainText
from .rsaalgorithmtype import RSAAlgorithmType
from .rsaoperation import RSAOperation
from .servermetadata import ServerMetadata
from .signer import Signer
from .signingoperation import SigningOperation
from .trustissues import *
from .undecryptable import Undecryptable
from .verifier import Verifier


__all__: list[str] = [
    'Algorithm',
    'AsymmetricSigningOperation',
    'AuthorizationServerMisbehaves',
    'AuthorizationServerNotDiscoverable',
    'AESAlgorithmType',
    'AESKeywrapOperation',
    'AESOperation',
    'AudienceType',
    'CipherText',
    'ClaimSet',
    'ClaimTypeError',
    'CryptographicFailure',
    'CryptographyPublicKeyType',
    'Data',
    'Decrypter',
    'Digest',
    'EdDSAOperation',
    'EdwardsCurveAlgorithmType',
    'EdwardsCurveType',
    'EllipticCurveAlgorithmType',
    'EllipticCurveType',
    'EllipticCurveOperation',
    'Encrypter',
    'EncryptOperation',
    'HMACAlgorithmType',
    'HMACOperation',
    'ForbiddenOperation',
    'GenerateAES',
    'GenerateEllipticCurve',
    'GenerateHMAC',
    'GenerateKeyOperation',
    'GenerateKeySpec',
    'GenerateOKP',
    'GenerateRSA',
    'IContentEncryptionKey',
    'IHasher',
    'IHTTPClient',
    'IKeychain',
    'IKeyInspector',
    'IKeySpecification',
    'InvalidCredential',
    'InvalidSignature',
    'IOperationPerformer',
    'IProvider',
    'JSONWebKey',
    'JSONWebKeySet',
    'JSONWebToken',
    'KeywrapOperation',
    'KeyAlgorithmType',
    'KeyOperationType',
    'KeyType',
    'KeyUseType',
    'Malformed',
    'MalformedHeader',
    'MalformedObject',
    'MalformedPayload',
    'MaximumAgeExceeded',
    'MissingProtectedClaim',
    'MissingProtectedHeader',
    'MissingRequiredClaim',
    'Message',
    'Operation',
    'PlainText',
    'RSAAlgorithmType',
    'RSAOperation',
    'ServerMetadata',
    'Signer',
    'SigningOperation',
    'TokenExpired',
    'TokenNotEffective',
    'TrustIssues',
    'Undecryptable',
    'UnknownAlgorithm',
    'UnresolvableIssuer',
    'UntrustedIssuer',
    'UntrustedSigningKey',
    'Verifier',
    'WrongAudience',
]


EncryptResult: TypeAlias = CipherText | Awaitable[CipherText]


# Copyright 2022 Cochise Ruhulessin
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import functools
from typing import cast
from typing import Any
from typing import Sequence

from pyasn1.type import univ
from pyasn1_modules import rfc5280

from ckms.ext import asn1
from ..oid import ID_CE_BASICCONSTRAINTS
from ..oid import ID_CE_KEYUSAGE
from ..oid import ID_CE_SUBJECTALTNAME
from ..oid import ID_EXTENSION_REQUEST
from .attribute import Attribute
from .certificateextension import BaseCertificateExtension
from .certificateextension import BasicConstraints
from .certificateextension import CertificateExtension
from .certificateextension import KeyUsage
from .certificateextension import SubjectAltNames
from .name import Name


class CertificateExtensionSequence(list[BaseCertificateExtension]):
    _index: dict[str, int]

    @property
    def basic_constraints(self) -> BasicConstraints:
        return cast(BasicConstraints, self.get(ID_CE_BASICCONSTRAINTS))

    @property
    def key_usage(self) -> KeyUsage:
        return cast(KeyUsage, self.get(ID_CE_KEYUSAGE))

    @property
    def subject_altnames(self) -> SubjectAltNames | None:
        return cast(SubjectAltNames, self.get(ID_CE_SUBJECTALTNAME))

    @classmethod
    def from_attribute(
        cls,
        attribute: Attribute
    ) -> 'CertificateExtensionSequence':
        extensions = cls([])
        obj, *_ = attribute.decode(rfc5280.Extensions())
        if _:
            raise ValueError("Unexpected length for Extensions.")
        for value in obj: # type: ignore
            ext = CertificateExtension.parse_obj({
                **value,
                'extnID': str(value[0]) # type: ignore
            })
            extensions.add(ext)
        return extensions

    def __init__(
        self,
        extensions: Sequence[BaseCertificateExtension],
        *args: Any,
        **kwargs: Any
    ):
        self._index = {}
        super().__init__(extensions, *args, **kwargs)
        for i, ext in enumerate(self):
            self._index[str(ext.ext_id)] = i
        if ID_CE_BASICCONSTRAINTS not in self._index:
            self.add(BasicConstraints.parse_obj({}))

    def add_san(self, san: Name) -> None:
        """A the given :class:`~ckms.ext.pkix.types.Name` to the Subject
        Alternative Name (SAN) extension.
        """
        if not self.subject_altnames:
            self.add(SubjectAltNames.parse_obj({}))
        assert self.subject_altnames is not None
        if san in self.subject_altnames.names:
            raise ValueError(
                f"Name {san} already in the Subject Alternative Name (SAN)"
            )
        self.subject_altnames.names.append(san)

    def as_attribute(self) -> Attribute:
        ext = rfc5280.Extensions()
        ext.extend([x.asn1() for x in self]) # type: ignore
        values = univ.SetOf()
        values.extend([ # type: ignore
            rfc5280.AttributeValue(asn1.encode_der(ext)) # type: ignore
        ])
        return Attribute(
            type=asn1.fields.ObjectIdentifier(ID_EXTENSION_REQUEST),
            values=values
        )

    def add(self, ext: BaseCertificateExtension) -> None:
        if str(ext.ext_id) in self._index:
            i = self._index[ext.ext_id]
            f = functools.partial(self.__setitem__, i)
        else:
            f = self.append
            i = len(self)
            self._index[ext.ext_id] = i
        return f(ext)

    def can_sign(self, value: bool | None = None) -> bool:
        """Return a boolean indicating if the extension allows the
        verification of digital signatures.
        """
        return self._keyusage(0, value)

    def get(self, oid: str) -> BaseCertificateExtension | None:
        attr: BaseCertificateExtension | None = None
        if oid in self._index:
            attr = self[self._index[oid]] # type: ignore
        return attr

    def nonrepudiable(self, value: bool | None = None) -> bool:
        """Return a boolean indicating if signatures made by the private
        key associated to the certificate public key are nonrepudiable.
        """
        return self._keyusage(1, value)

    def _keyusage(self, use: int, value: bool | None) -> bool:
        if value is not None and not self.key_usage:
            self.add(KeyUsage())
        assert self.key_usage is not None
        if value is not None:
            self.key_usage.set(use, int(value))
        return self.key_usage.get(use)


# pylint: skip-file
# Copyright 2018 Cochise Ruhulessin
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from unimatrix.exceptions import CanonicalException


__all__: list[str] = [
    'CryptographicFailure',
    'UnknownAlgorithm',
]


class CryptographicFailure(CanonicalException):
    __module__: str = 'ckms.types'
    http_status_code: int = 500
    code: str = 'CRYPTOGRAPHIC_FAILURE'
    message: str = (
        "An error occurred while performing a cryptographic operation. "
        "That's all we know."
    )


class UnknownAlgorithm(CryptographicFailure):
    __module__: str = 'ckms.types'
    algorithm: str
    http_status_code: int = 403
    code: str = "UNKNOWN_CRYPTOGRAPHIC_ALGORITHM"
    message: str = (
        "The cryptographic algorithm is not supported by the application. "
        "Consult the documentation for the list of supported algorithms."
    )

    def __init__(self, name: str):
        self.algorithm = name
        super().__init__()

    def __repr__(self) -> str:
        return f'<{type(self).__name__}: {self.algorithm} (id: {self.id})>'


# Copyright 2022 Cochise Ruhulessin
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from typing import Any

import pytest

from ckms.jose import PayloadCodec
from ckms.types import JSONWebToken


@pytest.mark.parametrize("signers", [
    ['HS256'],
    ['HS256', 'HS384'],
])
@pytest.mark.asyncio
async def test_jwt_serialization_is_valid(
    codec: PayloadCodec,
    signers: list[str]
):
    t1 = await codec.encode({'foo': 'bar'}, signers=signers)
    j1 = await codec.jws(t1)
    t2 = j1.json()
    j2 = await codec.jws(t2)
    jwt = await codec.decode(t2, verify=True)
    assert isinstance(jwt, JSONWebToken)
    assert j1.signatures[0] == j2.signatures[0]


@pytest.mark.parametrize("payload", [
    'Hello world!',
    b'Hello world!'
])
@pytest.mark.parametrize("signers", [
    ['HS256'],
    ['HS256', 'HS384'],
])
@pytest.mark.asyncio
async def test_jws_serialization_is_valid(
    codec: PayloadCodec,
    signers: list[str],
    payload: Any
):
    t1 = await codec.encode(payload, signers=signers)
    j1 = await codec.jws(t1)
    t2 = j1.json()
    j2 = await codec.jws(t2)
    jwt = await codec.decode(t2, verify=True)
    assert isinstance(jwt, bytes)
    assert j1.signatures[0] == j2.signatures[0]


# Copyright 2022 Cochise Ruhulessin
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from ckms.ext.asn1 import ObjectIdentifierEnum


class ExtendedKeyUsageType(str, ObjectIdentifierEnum):
    serverAuth          = '1.3.6.1.5.5.7.3.1'
    clientAuth          = '1.3.6.1.5.5.7.3.2'
    codeSigning         = '1.3.6.1.5.5.7.3.3'
    emailProtection     = '1.3.6.1.5.5.7.3.4'
    ipSecEndSystem      = '1.3.6.1.5.5.7.3.5'
    ipSecTunnel         = '1.3.6.1.5.5.7.3.6'
    ipSecUser           = '1.3.6.1.5.5.7.3.7'
    timeStamping        = '1.3.6.1.5.5.7.3.8'
    ocspSigning         = '1.3.6.1.5.5.7.3.9'
    secureShellClient   = '1.3.6.1.5.5.7.3.21'
    secureShellServer   = '1.3.6.1.5.5.7.3.22'


# Copyright 2022 Cochise Ruhulessin
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from unimatrix.exceptions import CanonicalException


__all__: list[str] = [
    'Malformed',
    'MissingProtectedClaim',
    'MissingProtectedHeader'
]


class Malformed(CanonicalException):
    __module__: str = 'ckms.core.types'
    http_status_code: int = 403


class MissingProtectedHeader(Malformed):
    http_status_code: int = 403
    code: str = 'MISSING_PROTECTED_HEADER'
    message: str = (
        "The JSON Web Signature (JWS) did not include a protected header "
        "for one or more signatures."
    )


class MissingProtectedClaim(Malformed):
    claim: str

    def __init__(self, name: str):
        self.claim = name
        super().__init__( # type: ignore
            message="A required claim was missing from the protected JOSE header.",
            detail=(
                f"The '{name}' claim was not included in the JOSE header and "
                "the application requires it to succcesfully validate the token."
            )
        )


# Copyright 2022 Cochise Ruhulessin
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Declares :class:`IProvider`."""
from typing import Any
from typing import TypeAlias
from typing import Union

from cryptography.hazmat.primitives.asymmetric import ec
from cryptography.hazmat.primitives.asymmetric import ed448
from cryptography.hazmat.primitives.asymmetric import ed25519
from cryptography.hazmat.primitives.asymmetric import rsa
from cryptography.hazmat.primitives.asymmetric import x448
from cryptography.hazmat.primitives.asymmetric import x25519

from .keyoperationtype import KeyOperationType


class IKeyInspector:
    __module__: str = 'cbra.types'

    PublicKeyType: TypeAlias = Union[
        ec.EllipticCurvePublicKey,
        ed448.Ed448PublicKey,
        ed25519.Ed25519PublicKey,
        rsa.RSAPublicKey,
        x448.X448PublicKey,
        x25519.X25519PublicKey
    ]

    PrivateKeyType: TypeAlias = Union[
        ec.EllipticCurvePrivateKey,
        ed448.Ed448PrivateKey,
        ed25519.Ed25519PrivateKey,
        rsa.RSAPrivateKey,
        x448.X448PrivateKey,
        x25519.X25519PrivateKey
    ]

    KeyType: TypeAlias = PublicKeyType | PrivateKeyType

    def calculate_kid(self, key: Any) -> str:
        raise NotImplementedError

    def from_jwk(self, jwk: dict[str, Any]) -> KeyType | bytes:
        raise NotImplementedError

    def get_algorithm_curve(self, algorithm: str) -> str | None:
        raise NotImplementedError

    def get_algorithm_use(self, algorithm: str) -> str:
        raise NotImplementedError

    def get_curve_by_algorithm(self, algorithm: str) -> ec.EllipticCurve:
        raise NotImplementedError

    def get_public_key_ops(self, private: list[KeyOperationType]) -> list[KeyOperationType]:
        raise NotImplementedError

    def to_bytes(self, key: Any) -> bytes:
        raise NotImplementedError

    def to_jwk(
        self,
        key: Any
    ) -> dict[str, Any]:
        """Return a dictionary holding a representation of the key as
        a JSON Web Key (JWK).
        """
        raise NotImplementedError


# Copyright 2022 Cochise Ruhulessin
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Declares :class:`KeySpecification`."""
from typing import Any
from typing import Generator

import dsnparse
import pydantic

from ckms.core import const
from ckms.types import Algorithm
from ckms.types import Decrypter
from ckms.types import Encrypter
from ckms.types import IKeyInspector
from ckms.types import IProvider
from ckms.types import JSONWebKey
from ckms.types import KeyOperationType
from ckms.types import KeyUseType
from ckms.types import Signer
from ckms.types import Verifier
from .basekeyspecification import BaseKeySpecification
from .certificate import Certificate
from .publickeyspecification import PublicKeySpecification


class KeySpecification(BaseKeySpecification, Decrypter, Encrypter, Signer, Verifier):
    """The base class for all provider key configurations."""
    __module__: str = 'ckms.core.models'
    _presets: dict[str, Any] = {}
    provider: IProvider
    kty: str
    kid: str | None = None
    use: KeyUseType | None
    algorithm: Algorithm = pydantic.Field(default=None, alias='alg')
    allow: set[str] = set()
    curve: str | None = None
    loaded: bool = False
    tags: set[str] = set()
    key_ops: set[KeyOperationType] = set()
    certificate: Certificate | None

    class Config(BaseKeySpecification.Config):
        allow_population_by_field_name: bool = True

    @pydantic.root_validator(allow_reuse=True, pre=True)
    def _autodiscover(cls, values: dict[str, Any]) -> dict[str, Any]:
        if values.get('alg'):
            values['algorithm'] = values['alg']
        cls.autodiscover(
            values['provider'],
            values['provider'].inspector,
            values
        )
        return values

    @classmethod
    def autodiscover(
        cls,
        provider: IProvider,
        inspector: IKeyInspector,
        values: dict[str, Any]
    ) -> None:
        """Hook to override key-specific validations."""
        pass

    @classmethod
    def parse_dsn(
        cls,
        dsn: dsnparse.ParseResult,
        **kwargs: Any
    ) -> dict[str, Any]:
        raise NotImplementedError

    def __init__(self, **kwargs: Any):
        super().__init__(**kwargs)
        assert isinstance(self.provider, IProvider)

    async def _load(self) -> 'KeySpecification':
        await self.load()
        if not self.key_ops:
            if self.algorithm in const.DH_ALGORITHMS:
                self.key_ops.add(KeyOperationType.deriveKey)
            if self.algorithm in const.ENCRYPTION_ALGORITHMS:
                self.key_ops.add(KeyOperationType.encrypt)
                if self.is_symmetric():
                    self.key_ops.add(KeyOperationType.decrypt)
            if self.algorithm in const.KEYWRAP_ALGORITHMS:
                self.key_ops.add(KeyOperationType.unwrapKey)
                if self.is_symmetric():
                    self.key_ops.add(KeyOperationType.wrapKey)
            if self.algorithm in const.SIGNING_ALGORITHMS:
                self.key_ops.add(KeyOperationType.sign)
                if self.is_symmetric():
                    self.key_ops.add(KeyOperationType.verify)
        if self.certificate is not None:
            await self.certificate.load()
        return self

    async def load(self) -> 'KeySpecification':
        raise NotImplementedError

    def as_public(self) -> PublicKeySpecification:
        ops: set[KeyOperationType]
        if not self.is_loaded():
            raise RuntimeError("Load the specification first before invoking this method.")
        params: dict[str, Any] = {
            'private': self,
            'provider': 'local',
            'kid': self.kid,
            'kty': self.kty,
            'alg': self.algorithm,
            'use': self.use,
            'crv': self.curve,
            'key': self.get_public_key(),
            'tags': self.tags,
            'certificate': self.certificate
        }
        params['key_ops'] = ops = set()
        if KeyOperationType.decrypt in self.key_ops:
            ops.add(KeyOperationType.encrypt)
        if KeyOperationType.deriveKey in self.key_ops:
            ops.add(KeyOperationType.deriveKey)
        if KeyOperationType.sign in self.key_ops:
            ops.add(KeyOperationType.verify)
        if KeyOperationType.unwrapKey in self.key_ops:
            ops.add(KeyOperationType.wrapKey)
        return PublicKeySpecification(**params)

    def as_jwk(self, private: bool = False) -> JSONWebKey:
        assert isinstance(self.provider, IProvider)
        if self.is_symmetric() and not private: # pragma: no cover
            raise TypeError(
                "Symmetric keys do not have a public key. Set `private=True` "
                "to project the private key to a JSON Web Key (JWK)."
            )
        claims: dict[str, Any] = {}
        if self.certificate:
            claims.update(self.certificate.claims())
        return self.provider.jwk(
            spec=self,
            private=private,
            **claims
        )

    def can_encrypt(self) -> bool:
        return self.can_perform(KeyOperationType.encrypt)

    def can_perform(self, op: str) -> bool:
        is_allowed = op in self.key_ops
        if self.is_asymmetric():
            public = self.as_public()
            is_allowed |= public.can_perform(op)
        return is_allowed

    def can_sign(self) -> bool:
        return self.can_perform(KeyOperationType.sign)

    def can_verify(self) -> bool:
        return self.use == 'sig'

    def can_wrap(self) -> bool:
        is_allowed = KeyOperationType.wrapKey in self.key_ops
        if self.is_asymmetric():
            is_allowed = KeyOperationType.unwrapKey in self.key_ops
        return is_allowed

    def get_key_material(self) -> bytes:
        """Return the key material, if available."""
        raise NotImplementedError

    def get_digest_oid(self) -> str | None:
        return self.algorithm.get_digest_oid()

    def is_aead(self) -> bool:
        return self.algorithm in {
            'A128GCM', 'A192GCM', 'A256GCM',
            'A128GCMKW', 'A192GCMKW', 'A256GCMKW'
        }

    def __await__(self) -> Generator[Any, None, 'KeySpecification']:
        if not self.is_loaded():
            return self._load().__await__()

        # TODO: a very quick and ugly hack.
        async def f():
            return self
        return f().__await__()



# Copyright 2018 Cochise Ruhulessin
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Declares :class:`JSONWebEncryption`."""
import inspect
from typing import cast
from typing import Any
from typing import Literal
from ckms.types.decrypter import Decrypter

import pydantic

from ckms.types import CipherText
from ckms.types import IKeychain
from ckms.types import IKeySpecification
from ckms.types import IProvider
from ckms.types import MalformedHeader
from ckms.types import MalformedPayload
from ckms.utils import b64decode_json
from .joseheader_ import JOSEHeader
from .octetb64 import OctetB64
from .recipient import Recipient


class JSONWebEncryption(pydantic.BaseModel):
    kind: Literal['JWE']
    token: str
    header: JOSEHeader
    protected: str = ''
    unprotected: dict[str, Any] = {}
    aad: str = ''
    ciphertext: OctetB64
    iv: OctetB64 | None
    tag: OctetB64
    recipients: list[Recipient]
    compact: bool = False

    @property
    def recipient(self) -> Recipient:
        """The first recipient in the array of recipients. Use as a
        shortcut for JWEs that have a single recipient.
        """
        return self.recipients[0]

    @pydantic.root_validator(allow_reuse=True, pre=True)
    def preprocess(cls, values: dict[str, Any]) -> dict[str, Any]:
        recipients: list[dict[str, Any]] = values.get('recipients') or []
        protected = b64decode_json(
            values.get('protected') or b'',
            on_failure=MalformedHeader(
                detail=(
                    "The JWE Protected Header could not be interpreted as "
                    "a Base64-encoded JSON object."
                )
            ),
            require=dict
        )
        unprotected = dict(values.get('unprotected') or {})

        assert isinstance(protected, dict)
        if set(protected.keys()) & set(unprotected.keys()):
            raise Exception
        common_header = values['header'] = JOSEHeader.parse_obj({
            **unprotected,
            **protected
        })
        for recipient in recipients:
            header = recipient.get('header') or {}
            if not isinstance(header, dict):
                raise MalformedHeader(
                    detail=(
                        "The per-recipient header must be a JSON object."
                    )
                )

            duplicates = set(header.keys()) & set(common_header.dict(exclude_defaults=True))
            if duplicates:
                raise MalformedHeader(
                    detail=(
                        "The claims in a per-recipient header must be "
                        "disjoint from the protected and unprotected "
                        "headers."
                    ),
                    hint=(
                        "The following claims were duplicate: "
                        f"{', '.join(sorted(duplicates))}."
                    )
                )
            recipient['header'] = common_header.clone(
                values=recipient.pop('header', {})
            )

        return values

    async def decrypt(self, keychain: IKeychain) -> bytes:
        """Decrypt the payload."""
        cek = self.decrypt_cek(keychain)
        if inspect.isawaitable(cek):
            cek = await cek
        return await cek.decrypt( # type: ignore
            CipherText(
                buf=bytes(self.ciphertext),
                aad=self.get_aad(),
                iv=bytes(self.iv or b''),
                tag=bytes(self.tag)
            )
        )

    async def decrypt_cek(self, keychain: IKeychain) -> IKeySpecification:
        """Return a :class:`~ckms.types.IKeySpecification` describing the
        key that was used to encrypt the payload.
        """
        kid = self.header.kid
        cek = None
        provider = IProvider.get('local')
        spec = None

        # If direct encryption is used and the kid claim is in the header,
        # immediately look up the key and bail out (TODO: audit).
        if keychain.has(kid) and self.is_direct():
            return cast(IKeySpecification, keychain.get(kid))

        # Find any recipient for which we can decrypt the CEK or find
        # the direct encryption key.
        for recipient in self.recipients:
            kid = recipient.kid or kid
            if kid is None or not keychain.has(kid):
                continue
            spec = cast(Decrypter, keychain.get(kid))
            if not self.is_direct():
                ct = CipherText(
                    buf=recipient.encrypted_key,
                    iv=recipient.iv,
                    tag=recipient.tag
                )
                cek = spec.decrypt(ct)
                if inspect.isawaitable(cek):
                    cek = await cek
                spec = provider.parse_spec({
                    'kty': 'oct',
                    'algorithm': self.header.enc,
                    'use': 'enc',
                    'key': {'cek': cek}
                })
                await spec
            break
        else:
            raise MalformedPayload(
                detail=(
                    "No key could be found to decrypt the payload."
                )
            )
        return spec

    def get_content_type(self) -> str | None:
        """Return the content type from the ``cty`` header."""
        return str.lower(self.header.cty or '') if self.header.cty else None

    def get_type(self) -> str | None:
        """Return the object type from the ``typ`` header."""
        return str.lower(self.header.typ or '') if self.header.typ else None

    def is_direct(self) -> bool:
        """Return a boolean indicating if the JWE was encrypted in Direct
        Encryption or Direct Key Agreement mode.
        """
        # TODO: We assume here that a JWE in Direct Encryption or
        # Direct Key Agreement has only one recipient.
        if len(self.recipients) != 1:
            return False
        return self.recipient.alg == 'dir'

    def get_aad(self) -> bytes | None:
        """Return the the Additional Authenticated Data (AAD) for AEAD
        encryption algorithms.
        """
        aad = str.encode(self.protected or '', 'ascii')
        #if self.aad:
        #    aad += b'.' + b64encode(self.aad)
        return aad

    def get_headers(self) -> list[JOSEHeader]:
        """Return the list of :class:`JOSEHeader` containing the aggregated
        header claims for all recipients.
        """
        return [x.header for x in self.recipients]


# Copyright 2022 Cochise Ruhulessin
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import copy
from typing import Any

import pytest_asyncio

from ckms.core import Keychain


@pytest_asyncio.fixture(scope='session') # type: ignore
async def keychain(keys: dict[str, Any]):
    keychain = Keychain()
    keychain.configure(keys=copy.deepcopy(keys))
    await keychain
    return keychain


@pytest_asyncio.fixture(scope='session', autouse=True) # type: ignore
async def setup_keychain(keychain: Keychain):
    await keychain


# Copyright 2022 Cochise Ruhulessin
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import copy
import enum
from typing import cast
from typing import Any

import pytest
from jwcrypto.common import json_encode # type: ignore

from ckms.core import const
from ckms.core import Keychain
from ckms.core.exceptions import MissingProtectedClaim
from ckms.core.models import JOSEObject
from ckms.core.models import JSONWebSignature
from ckms.core.models import KeySpecification
from .baseprovidertestcase import BaseProviderTestCase



class JOSELibraryCapability(str, enum.Enum):
    multisign = 'multisign'


class JOSEConsumerTestCase(BaseProviderTestCase):
    name: str = 'local'
    payload: bytes = b'Hello world!'
    capabilities: set[str] = set()
    Capability: type[enum.Enum] = JOSELibraryCapability

    def create_jws(
        self,
        signers: list[tuple[KeySpecification, dict[str, Any] | None, dict[str, Any] | None]],
        claims: dict[str, Any] | None = None,
        compact: bool = False,
        encode: bool = False
    ) -> bytes:
        """Create a JSON Web Signature (JWS) using the given claims."""
        raise NotImplementedError

    @pytest.mark.parametrize("algorithm", const.SIGNING_ALGORITHMS)
    @pytest.mark.parametrize("compact", [True, False])
    @pytest.mark.parametrize("encode", [True, False])
    @pytest.mark.parametrize("require_kid", [True, False])
    @pytest.mark.parametrize("protected", [
        {'foo': 1},
    ])
    @pytest.mark.parametrize("header", [
        None,
        {'bar': 1}
    ])
    @pytest.mark.asyncio
    async def test_sign_and_verify_jws_single(
        self,
        algorithm: str,
        keychain: Keychain,
        protected: dict[str, Any] | None,
        header: dict[str, Any] | None,
        compact: bool,
        encode: bool,
        require_kid: bool
    ):
        if header is not None:
            header = copy.deepcopy(header)
        spec = keychain.get(algorithm)
        protected = copy.deepcopy(protected or {})
        protected.setdefault('alg', spec.algorithm)
        assert spec.kid is not None
        if require_kid:
            header = header or {}
            header.setdefault('kid', spec.kid)
        if compact:
            protected = protected or {}
            protected.update(header or {})
            header = None

        buf = self.create_jws(
            signers=[
                (spec, protected, header)
            ],
            compact=compact,
            encode=encode
        )
        jws = cast(JSONWebSignature, JOSEObject.parse(buf))

        assert isinstance(jws, JSONWebSignature)
        success = await jws.verify(keychain, require_kid=require_kid)
        if protected.get('alg') != spec.algorithm:
            assert not success
        else:
            assert success

    @pytest.mark.parametrize("algorithm", const.SIGNING_ALGORITHMS)
    @pytest.mark.asyncio
    async def test_sign_with_missing_kid_never_validates(
        self,
        algorithm: str,
        keychain: Keychain    
    ):
        spec = keychain.get(algorithm)
        buf = self.create_jws(
            signers=[
                (spec, {'alg': spec.algorithm}, None)
            ]
        )
        jws = cast(JSONWebSignature, JOSEObject.parse(buf))
        try:
            assert not await jws.verify(keychain, require_kid=True)
        except MissingProtectedClaim as exception:
            assert exception.claim == 'kid'

    @pytest.mark.parametrize("algorithm", const.SIGNING_ALGORITHMS)
    @pytest.mark.asyncio
    async def test_sign_with_missing_kid_validates_if_require_kid_is_false(
        self,
        algorithm: str,
        keychain: Keychain    
    ):
        spec = keychain.get(algorithm)
        buf = self.create_jws(
            signers=[
                (spec, {'alg': spec.algorithm}, None)
            ]
        )
        jws = cast(JSONWebSignature, JOSEObject.parse(buf))
        assert await jws.verify(keychain, require_kid=False)


# pylint: skip-file
import os
import typing

import pytest
from cryptography.hazmat.primitives.ciphers import algorithms
from cryptography.hazmat.primitives.asymmetric import ec
from cryptography.hazmat.primitives.asymmetric import ed448
from cryptography.hazmat.primitives.asymmetric import x448
from cryptography.hazmat.primitives.asymmetric import ed25519
from cryptography.hazmat.primitives.asymmetric import x25519
from cryptography.hazmat.primitives.asymmetric import rsa

from ckms.core import KeyInspector


ASYMMETRIC_KEYS: list[typing.Any] = [
    ec.generate_private_key(curve=ec.SECP256R1()),
    ec.generate_private_key(curve=ec.SECP256R1()).public_key(),
    ed448.Ed448PrivateKey.generate(),
    ed448.Ed448PrivateKey.generate().public_key(),
    ed25519.Ed25519PrivateKey.generate(),
    ed25519.Ed25519PrivateKey.generate().public_key(),
    x448.X448PrivateKey.generate(),
    x448.X448PrivateKey.generate().public_key(),
    x25519.X25519PrivateKey.generate(),
    x25519.X25519PrivateKey.generate().public_key(),
    rsa.generate_private_key(public_exponent=65537, key_size=1024),
    rsa.generate_private_key(public_exponent=65537, key_size=1024).public_key(),
]


@pytest.mark.parametrize("key", ASYMMETRIC_KEYS + [
    algorithms.AES(os.urandom(16)),
    algorithms.AES(os.urandom(24)),
    algorithms.AES(os.urandom(32)),
    'foo'
])
def test_keyinspector_supports_keytypes(key: KeyInspector.KeyType | str):
    inspector = KeyInspector()
    inspector.calculate_kid(key)


@pytest.mark.parametrize("key", [
    ec.generate_private_key(curve=ec.SECP256R1()),
    ed448.Ed448PrivateKey.generate(),
    ed25519.Ed25519PrivateKey.generate(),
    x448.X448PrivateKey.generate(),
    x25519.X25519PrivateKey.generate(),
    rsa.generate_private_key(public_exponent=65537, key_size=1024),
])
def test_private_kid_equals_public_kid(key: KeyInspector.PrivateKeyType):
    inspector = KeyInspector()
    public = key.public_key()
    assert inspector.calculate_kid(key) == inspector.calculate_kid(public)


@pytest.mark.parametrize("key", ASYMMETRIC_KEYS)
def test_to_pem(key: KeyInspector.KeyType):
    inspector = KeyInspector()
    pem = inspector.to_pem(key)
    assert isinstance(pem, bytes)

@pytest.mark.parametrize("key", ASYMMETRIC_KEYS + [
    algorithms.AES(os.urandom(16)),
    algorithms.AES(os.urandom(24)),
    algorithms.AES(os.urandom(32)),
    os.urandom(64)
])
def test_to_jwk(key: KeyInspector.KeyType | bytes):
    inspector = KeyInspector()
    inspector.to_jwk(key)


@pytest.mark.parametrize("key", ASYMMETRIC_KEYS + [
    algorithms.AES(os.urandom(16)),
    algorithms.AES(os.urandom(24)),
    algorithms.AES(os.urandom(32)),
    os.urandom(64)
])
def test_from_jwk(key: KeyInspector.KeyType | bytes):
    inspector = KeyInspector()
    decoded = inspector.from_jwk(
        typing.cast(dict[str, str], inspector.to_jwk(key))
    )

    assert inspector.calculate_kid(decoded) == inspector.calculate_kid(key)
    if isinstance(key, algorithms.AES):
        key = key.key
    assert type(key) == type(decoded)


# pylint: skip-file
# Copyright 2018 Cochise Ruhulessin
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from typing import cast
from typing import Any

import pydantic
from cryptography.hazmat.primitives.asymmetric.types import PUBLIC_KEY_TYPES

from ckms.types import Algorithm
from ckms.types import JSONWebKey
from ckms.types import KeyOperationType
from ckms.types import KeyUseType
from ckms.core.models import KeySpecification
from ckms.utils import b64encode
from .types import AzureConsumer


class AzureKeySpecification(KeySpecification, AzureConsumer):
    algorithm: Algorithm = Algorithm.get('null')
    kty: str | None # type: ignore
    use: KeyUseType
    name: str
    vault: str | None
    vault_url: str
    jwk: JSONWebKey | None
    hsm: bool = False

    #: The key version. If :attr:`version` is ``None``, then the latest
    #: version is used.
    version: str | None

    @property
    def azure_key_id(self) -> str:
        return f'{self.vault_url}/keys/{self.name}/{self.version}'

    @pydantic.root_validator(pre=True)
    def preprocess(
        cls,
        values: dict[str, Any]
    ) -> dict[str, Any]:
        vault = values.get('vault')
        if vault and not values.get('vault_url'):
            values['vault_url'] = f"https://{vault}.vault.azure.net/"
        return values

    @pydantic.validator('vault_url', pre=False)
    def postprocess_vault_url(cls, value: str | None) -> str | None:
        if isinstance(value, str):
            value = str.rstrip(value, '/')
        return value

    def get_public_key(self) -> PUBLIC_KEY_TYPES | None:
        assert self.jwk is not None
        return self.jwk.get_public_key()

    async def load(self) -> KeySpecification:
        async with self.key_client(self.vault_url) as client:
            obj = await client.get_key(self.name, version=self.version)
            self.kty = str(obj.key_type)
            if self.kty in {'EC-HSM', 'RSA-HSM'}: # pragma: no cover
                self.kty, _ = str.split(self.kty, '-') # type: ignore
                self.hsm = True
            jwk = cast(JSONWebKey, obj.key)
            self.kid = self.provider.calculate_kid(obj.id)
            self.key_ops = {KeyOperationType(x) for x in obj.key_operations}
            if jwk.crv: # type: ignore
                self.algorithm = Algorithm.get_for_curve(jwk.crv)
                self.curve = jwk.crv
            else:
                assert self.use is not None
                if self.algorithm is None: # pragma: no cover
                    self.algorithm = Algorithm.default(self.kty, self.use.value)
            self.jwk = JSONWebKey.parse_obj({
                **{
                    k: bytes.decode(b64encode(v)) if isinstance(v, bytes) else v
                    for  k, v in jwk.__dict__.items()
                },
                'alg': self.algorithm,
                'use': self.use,
                'kid': self.kid,
                'kty': self.kty
            })
            if self.version is None:
                self.version = obj.properties.version

        assert self.algorithm
        assert self.kty
        assert self.use
        assert self.key_ops
        self.loaded = True
        return self

    def __hash__(self) -> int: # type: ignore
        return str.__hash__(f'{self.vault_url}/{self.name}')



# Copyright 2018 Cochise Ruhulessin
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Declares :class:`AESAlgorithmType`."""
import enum


class AESAlgorithmType(str, enum.Enum):
    A128KW          = 'A128KW'
    A192KW          = 'A192KW'
    A256KW          = 'A256KW'
    A128GCM         = 'A128GCM'
    A192GCM         = 'A192GCM'
    A256GCM         = 'A256GCM'
    A128GCMKW       = 'A128GCMKW'
    A192GCMKW       = 'A192GCMKW'
    A256GCMKW       = 'A256GCMKW'
    A128CBC_HS256   = 'A128CBC-HS256'
    A192CBC_HS384   = 'A192CBC-HS384'
    A256CBC_HS512   = 'A256CBC-HS512'


# Copyright 2018 Cochise Ruhulessin
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import functools
from typing import Any

from .keyoperationtype import KeyOperationType
from .forbiddenoperation import ForbiddenOperation


class IOperationPerformer:
    __module__: str = 'ckms.types'
    algorithm: Any
    tags: set[str]
    use: Any

    @staticmethod
    def must_allow(op: KeyOperationType) -> Any:
        def decorator_factory(func: Any) -> Any:
            @functools.wraps(func)
            def f(self: IOperationPerformer, *args: Any, **kwargs: Any) -> Any:
                if not self.can_perform(op) and not kwargs.pop('force', False):
                    raise ForbiddenOperation(op)
                return func(self, *args, **kwargs)
            return f
        return decorator_factory

    def allows_algorithm(self, algorithm: str) -> bool:
        """Return a boolean indicating if the given algorithm may be used with
        the key.
        """
        return self.algorithm == algorithm

    def as_public(self) -> Any:
        raise NotImplementedError

    def can_encrypt(self) -> bool:
        """Return a boolean indicating if this implementation can encrypt
        data.
        """
        raise NotImplementedError

    def can_sign(self) -> bool:
        """Return a boolean indicating if this implementation can create
        a digital signature.
        """
        raise NotImplementedError

    def can_perform(self, op: str | KeyOperationType) -> bool:
        """Return a boolean indicating if the key can perform the
        given cryptographic operation.
        """
        raise NotImplementedError

    def can_verify(self) -> bool:
        """Return a boolean indicating if this implementation can verify
        a digital signature.
        """
        raise NotImplementedError

    def has_key_material(self) -> bool:
        """Return a boolean indicating if the key material"""
        raise NotImplementedError

    def is_asymmetric(self) -> bool:
        """Return a boolean indicating if the key is an asymmetric
        algorithm.
        """
        raise NotImplementedError

    def is_loaded(self) -> bool: # pragma: no cover
        """Return a boolean indicating if the metadata describing the
        key is retrieved.
        """
        raise NotImplementedError

    def is_public(self) -> bool:
        """Return a boolean indicating if this specification represent
        a public key.
        """
        raise NotImplementedError


# Copyright 2018 Cochise Ruhulessin
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# type: ignore
from typing import Any

import pydantic
from cryptography import x509
from cryptography.hazmat.primitives.asymmetric import rsa
from cryptography.hazmat.primitives.asymmetric import ec
from pyasn1.type.base import Asn1Type
from pyasn1_modules import rfc2315
from pyasn1.codec.ber import decoder

from ckms.ext import asn1
from ckms.ext.pkcs.oid import SIGNED_DATA
from ckms.ext.pkcs.types import SignedData
from .timestampreq import TimeStampReq
from .tstinfo import TSTInfo


class TimeStampToken(pydantic.BaseModel):
    message: bytes
    signature: SignedData
    info: TSTInfo

    @pydantic.root_validator(pre=True)
    def preprocess(cls, values: dict[str, Any]) -> dict[str, Any]:
        if str(values.get('contentType')) != SIGNED_DATA:
            raise ValueError("Not a SignedData object")
        values['content_type'] = str(values['contentType'])
        if isinstance(values.get('content'), Asn1Type):
            obj, _ = decoder.decode(values['content'], asn1Spec=rfc2315.SignedData())
            if _:
                raise ValueError("Invalid BER-encoding")
            values['signature'] = obj
        values['message'] = bytes(asn1.decode_ber(bytes(obj['contentInfo']['content'])))
        values['info'] = TSTInfo.from_signed_data(obj['contentInfo'])
        return values

    def verify(
        self,
        certificate: x509.Certificate | None = None,
        request: TimeStampReq | None = None
    ) -> bool:
        """Verify the time stamp token using the given public key.
        
        Verify that the signer produced a signature of the correct message
        digest. The :meth:`verify()` method assumes that there is one signer.
        """
        s = self.signature.signers[0]
        if s.signed_digest != s.get_digest(bytes(self.message)):
            raise ValueError("Digest mismatch for signed data.")
        if certificate is not None:
            k: ec.EllipticCurvePublicKey | rsa.RSAPublicKey = certificate.public_key()
            s.verify(k, self.signature.content_info.content)
        if request is not None:
            if request.nonce and not self.info.nonce:
                raise ValueError("TSA did not return the nonce.")
            if request.nonce != self.info.nonce:
                raise ValueError("TSA returned a different nonce.")
        return True



# Copyright 2022 Cochise Ruhulessin
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import copy
import json
import types
from typing import Any

import pytest

from ckms.core import Keychain
from ckms.core.const import ENCRYPTION_ALGORITHMS
from ckms.core.models import KeySpecification
from ckms.jose.decoder import Decoder
from ckms.jose.encoder import Encoder
from ckms.utils import b64encode_json
from .josebasetestcase import JOSEBaseTestCase


TEST_PAYLOADS: list[bytes | dict[str, Any]] = [
    b'Hello world!',
    str.encode(json.dumps({'foo': 'bar'}), 'utf-8')
]


class JOSEEncryptionTestCase(JOSEBaseTestCase):
    __module__: str = 'ckms.core.provider.tests'
    supported_encryption_algorithms: set[str] = set(ENCRYPTION_ALGORITHMS)
    supported_wrapping_algorithms: set[str] = set()
    tests_package: types.ModuleType | None = None

    async def create_jwe(
        self,
        *,
        encoder: Encoder,
        recipients: list[tuple[KeySpecification, dict[str, Any] | None]],
        recipient: KeySpecification |  None = None,
        payload: bytes | dict[str, Any],
        signers: list[Any] | None = None,
        expect_jwt: bool = False,
        compact: bool = False,
        direct: bool = False
    ) -> str:
        signers = signers or []
        obj = encoder.encode(content=payload)
        for spec, protected, header in signers:
            obj.add_signature(signer=spec, protected=protected, header=header)
        for spec, header in recipients:
            obj.add_recipient(encrypter=spec, header=header, direct=direct)
        return await obj.serialize(compact=compact)

    async def decrypt_jwe(
        self,
        encoder: Encoder,
        decoder: Decoder,
        recipients: list[tuple[KeySpecification, dict[str, Any] | None]],
        signers: list[list[Any]],
        payload: bytes | dict[str, Any],
        token: str
    ) -> Any:
        return await decoder.decode(token)

    @pytest.mark.parametrize("name", ["RSA-OAEP"])
    @pytest.mark.parametrize("encryption", ["A256GCM"])
    @pytest.mark.parametrize("payload", [
        b'Hello world!',
        {'foo': 'bar'}
    ])
    @pytest.mark.parametrize("header", [
        None,
        {'foo': 'bar'}
    ])
    @pytest.mark.parametrize("compact", [True, False])
    @pytest.mark.parametrize("signers", [
        None,
        [('HS256', None, None)],
        [('HS256', None, None), ('HS384', None, None)],
    ])
    @pytest.mark.asyncio
    async def test_wrap_single(
        self,
        encoder: Encoder,
        decoder: Decoder,
        keychain: Keychain,
        name: str,
        encryption: str | None,
        payload: bytes | dict[str, Any],
        header: dict[str, Any] | None,
        compact: bool,
        signers: list[list[Any]],
        direct: bool = False,
        recipients: list[tuple[KeySpecification, dict[str, Any] | None]] | None = None
    ):
        is_jwt = isinstance(payload, dict)
        payload = copy.deepcopy(payload)
        signers = copy.deepcopy(list(signers or []))
        if not recipients:
            spec: KeySpecification = keychain.get(name)
            if direct and not spec.can_encrypt():
                pytest.skip()
            elif not direct and not spec.can_wrap():
                pytest.skip()
            elif not direct and spec.algorithm\
            not in self.supported_wrapping_algorithms:
                pytest.skip()
            recipients = [
                (spec, header)
            ]
    
        for i, signer in enumerate(signers):
            signer_name, *headers = signer
            signers[i] = [keychain.get(signer_name)] + headers

        compact = compact and (len(recipients) <= 1 and len(signers) <= 1)
        token = await self.create_jwe(
            encoder=encoder,
            recipients=recipients,
            payload=payload,
            compact=compact,
            direct=direct,
            signers=signers
        )
        if isinstance(token, dict):
            token = bytes.decode(b64encode_json(token), 'ascii')
        await self.decrypt_jwe(
            encoder=encoder,
            decoder=decoder,
            recipients=recipients,
            signers=signers,
            payload=payload,
            token=token
        )

        headers, _ = decoder.introspect(token)
        typ = "JOSE" if compact else "JOSE+JSON"
        cty = "application/octet-stream"
        if recipients and signers:
            cty = "JOSE" if compact else "JOSE+JSON"
        if is_jwt:
            cty = None
            typ = "JWT"
            if signers:
                cty = "JWT"
        for i, jose in enumerate(headers):
            assert jose.alg in {recipients[i][0].algorithm, 'dir'}
            assert jose.kid == recipients[i][0].kid
            assert jose.typ == typ
            assert jose.cty == cty, jose

    @pytest.mark.parametrize("recipient1", [
        ('A128GCMKW', None),
        ('A128GCMKW', {'foo': 'bar'}),
    ])
    @pytest.mark.parametrize("recipient2", [
        ('A192GCMKW', None),
        ('A192GCMKW', {'foo': 'bar'}),
    ])
    @pytest.mark.parametrize("encryption", sorted(ENCRYPTION_ALGORITHMS))
    @pytest.mark.parametrize("payload", [
        b'Hello world!',
        str.encode(json.dumps({'foo': 'bar'}), 'utf-8')
    ])
    @pytest.mark.asyncio
    async def test_wrap_multi(
        self,
        encoder: Encoder,
        decoder: Decoder,
        keychain: Keychain,
        recipient1: tuple[str, dict[str, Any] | None],
        recipient2: tuple[str, dict[str, Any] | None],
        encryption: str,
        payload: bytes
    ):
        recipients = [
            (keychain.get(recipient1[0]), recipient1[1]),
            (keychain.get(recipient2[0]), recipient2[1]),
        ]
        await self.test_wrap_single(
            encoder=encoder,
            decoder=decoder,
            keychain=keychain,
            name='',
            encryption=encryption,
            payload=payload,
            header={},
            compact=False,
            recipients=recipients,
            signers=[]
        )

    @pytest.mark.parametrize("compact", [True, False])
    @pytest.mark.parametrize("header", [
        None,
        {'foo': 'bar'}
    ])
    @pytest.mark.parametrize("name", sorted(ENCRYPTION_ALGORITHMS))
    @pytest.mark.parametrize("payload", [
        b'Hello world!',
        str.encode(json.dumps({'foo': 'bar'}), 'utf-8')
    ])
    @pytest.mark.asyncio
    async def test_direct_single(
        self,
        encoder: Encoder,
        decoder: Decoder,
        keychain: Keychain,
        name: str,
        payload: bytes | dict[str, Any],
        header: dict[str, Any] | None,
        compact: bool
    ):
        await self.test_wrap_single(
            encoder=encoder,
            decoder=decoder,
            keychain=keychain,
            name=name,
            encryption=None,
            payload=payload,
            header=header,
            compact=compact,
            direct=True,
            signers=[]
        )


# Copyright 2022 Cochise Ruhulessin
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from typing import Any
from typing import Literal
from typing import TypeAlias

from cryptography.hazmat.primitives.asymmetric import ed448
from cryptography.hazmat.primitives.asymmetric import ed25519
from cryptography.hazmat.primitives.asymmetric import x448
from cryptography.hazmat.primitives.asymmetric import x25519

from ckms.core import const
from ckms.types import EdwardsCurveType
from ckms.types import IKeyInspector
from ckms.types import IProvider
from ckms.types import KeyUseType
from .localkey import LocalKey
from .localkeyspecification import LocalKeySpecification
from .parameterlesskey import ParameterLessKey


OKPPublic: TypeAlias = (
    ed448.Ed448PublicKey |
    ed25519.Ed25519PublicKey |
    x448.X448PublicKey |
    x25519.X25519PublicKey
)


OKPPrivate: TypeAlias = (
    ed448.Ed448PrivateKey |
    ed25519.Ed25519PrivateKey |
    x448.X448PrivateKey |
    x25519.X25519PrivateKey
)


class EdwardsCurveKeySpecification(LocalKeySpecification):
    kty: Literal['OKP'] = 'OKP'
    #algorithm: types.EdwardsCurveAlgorithmType | None
    curve: EdwardsCurveType | None
    key: LocalKey | ParameterLessKey = ParameterLessKey()
    use: KeyUseType

    @classmethod
    def autodiscover(
        cls,
        provider: IProvider,
        inspector: IKeyInspector,
        values: dict[str, Any]
    ) -> None:
        if not values.get('use') and not values.get('curve'):
            raise ValueError(
                "Specify either the `use` parameter the `curve` parameter."
            )
        if values.get('curve') in {'Ed448', 'Ed25519'}:
            values['use'] = 'sig'
        elif values.get('curve') in {'X448', 'X25519'}:
            values['use'] = 'enc'
        elif values.get('use') == 'sig':
            values['curve'] = const.DEFAULT_ED_SIGNING_CURVE
        elif values.get('use') == 'enc':
            values['curve'] = const.DEFAULT_ED_ENCRYPTION_CURVE
        else:
            raise ValueError("Unable to infer key parameters.")
        if values['use'] == 'sig':
            values['algorithm'] = 'EdDSA'
        if values['use'] == 'enc' and not values.get('algorithm'):
            values['algorithm'] = const.DEFAULT_ED_ENCRYPTION_ALGORITHM

    def get_public_key(self) -> OKPPublic:
        assert isinstance(self.key, OKPPrivate)
        return self.key.public_key()


# Copyright 2022 Cochise Ruhulessin
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import enum
import functools
import pathlib
from typing import Any

import pydantic
from cryptography import x509
from cryptography.hazmat.primitives.asymmetric.types import PUBLIC_KEY_TYPES
from cryptography.hazmat.primitives.serialization import load_pem_public_key
from cryptography.hazmat.primitives.serialization import load_pem_private_key

from ckms.utils import b64decode
from .algorithm import Algorithm
from .encrypter import Encrypter
from .iprovider import IProvider
from .ikeyinspector import IKeyInspector
from .ikeyspecification import IKeySpecification
from .keyoperationtype import KeyOperationType
from .keyusetype import KeyUseType
from .signer import Signer
from .verifier import Verifier


class JSONWebKey(Encrypter, Verifier, Signer, pydantic.BaseModel, IKeySpecification):
    __module__: str = 'ckms.types'

    @property # type: ignore pragma: no cover
    def algorithm(self) -> str | None:
        return self.alg

    @property
    def inspector(self) -> IKeyInspector:
        return self.provider.inspector # type: ignore

    @property # type: ignore
    def provider(self) -> IProvider:
        return IProvider.get('local')

    kty: str | None
    kid: str | None
    alg: Algorithm | None
    crv: str | None
    use: KeyUseType | None # type: ignore
    key_ops: list[KeyOperationType] = []
    x5u: str | None
    x5c: list[str] | None
    x5t: str | None
    x5t_sha256: str | None = pydantic.Field(
        default=None,
        alias="x5t#S256"
    )

    # Key material
    d: str | None
    e: str | None
    n: str | None
    k: str | None
    p: str | None
    q: str | None
    x: str | None
    y: str | None
    dp: str | None
    dq: str | None
    qi: str | None

    @classmethod
    def fromcertificate(cls, *, alg: str, crt: x509.Certificate, **claims: Any) -> 'JSONWebKey':
        """Create a new :class:`JSONWebKey` from a :class:`cryptography.x509.Certificate`
        instance.
        """
        p = IProvider.get('local')
        return cls.parse_obj({
            **p.inspector.to_jwk(crt.public_key()),
            **claims,
            'alg': alg,
            # TODO: Discover key ops from certificate extensions
            'key_ops': {'verify', 'encrypt', 'wrapKey'}
        })

    @classmethod
    def fromfile(cls, filepath: str | pathlib.Path, **claims: Any) -> 'JSONWebKey':
        """Create a new :class:`JSONWebKey` from a PEM-encoded key or
        X.509 certificate at the given `filepath`.
        """
        with open(filepath, 'rb') as f:
            pem = f.read()
            if b'-----BEGIN CERTIFICATE-----' in pem:
                return cls.fromcertificate(
                    crt=x509.load_pem_x509_certificate(pem),
                    **claims
                )
            else:
                return cls.frompem(pem=pem, **claims)

    @classmethod
    def frompem(
        cls,
        pem: bytes,
        *,
        alg: str | None = None,
        use: str | KeyUseType | None = None,
        key_ops: list[str] | list[KeyOperationType] | None = None,
        password: bytes | str | None = None,
        public: bool = False,
        **claims: Any
    ) -> 'JSONWebKey':
        """Create a new :class:`JSONWebKey` instance from a PEM-encoded
        key.
        """
        p = IProvider.get('local')
        f = load_pem_public_key
        if b'PRIVATE KEY-----' in pem:
            if isinstance(password, str): # pragma: no cover
                password = str.encode(password)
            f = functools.partial(load_pem_private_key, password=password)
        jwk = cls.parse_obj({
            'alg': alg,
            'use': use,
            **p.inspector.to_jwk(f(pem)),
            **claims,
            'key_ops': key_ops or [],
        })
        if public and jwk.is_asymmetric():
            jwk = jwk.as_public()
        return jwk

    @classmethod
    def frompublic(cls, key: PUBLIC_KEY_TYPES, **claims: Any) -> 'JSONWebKey':
        p = IProvider.get('local')
        return cls.frompem(p.inspector.to_bytes(key), **claims)

    @pydantic.root_validator(allow_reuse=True, pre=True)
    def preprocess(cls, values: dict[str, Any]) -> dict[str, Any]:
        kty = values.get('kty')
        if kty in {'EC', 'OKP'} and not values.get('crv'): # pragma: no cover
            raise ValueError(
                "The 'crv' parameter is required for keys of type "
                "EC or OKP."
            )
        alg = values.get('alg')
        if isinstance(alg, enum.Enum):
            alg = alg.value
        if alg is not None and not isinstance(alg, Algorithm):
            alg = Algorithm.get(alg)
        if alg is not None:
            values['alg'] = alg
            values['use'] = alg.use
        if not values.get('alg') and not values.get('use'):
            raise TypeError("Provide either the 'alg' or 'use' parameter.")
        return values

    def __init__(self, **kwargs: Any) -> None:
        super().__init__(**kwargs)
        if self.kid is None:
            if self.is_asymmetric():
                self.kid = self.inspector.calculate_kid(self.get_public_key())
            elif self.k:
                self.kid = self.inspector.calculate_kid(b64decode(self.k))
        if self.alg is None:
            assert self.kty
            assert self.use
            self.alg = Algorithm.default(self.kty, self.use)

        assert self.alg is not None
        if not self.key_ops:
            self.key_ops = [
                getattr(KeyOperationType, op)
                for op in self.alg.get_key_ops(self.is_public())
            ]

    def as_public(self) -> 'JSONWebKey':
        if not self.is_public():
            key = self.inspector.from_jwk(jwk=self.dict())
            self = JSONWebKey.parse_obj({
                **self.inspector.to_jwk(key.public_key()), # type: ignore
                'kid': self.kid,
                'alg': self.alg,
                'use': self.use,
                'key_ops': self.inspector.get_public_key_ops(self.key_ops),
                'x5u': self.x5u,
                'x5c': self.x5c,
                'x5t': self.x5t,
                'x5t#S256': self.x5t_sha256
            })
        return self

    def can_encrypt(self) -> bool:
        return bool(set(self.key_ops) & {'encrypt', KeyOperationType.encrypt})

    def can_perform(self, op: KeyOperationType | str) -> bool: # pragma: no cover
        return op in self.key_ops

    def can_verify(self) -> bool:
        return KeyOperationType.verify in self.key_ops

    def dict(self, **kwargs: Any) -> dict[str, Any]: # type: ignore pragma: no cover
        kwargs.setdefault('exclude_defaults', True)
        return super().dict(**kwargs)

    def get_public_key(self) -> Any:
        if not self.is_public():
            self = self.as_public()
        return self.inspector.from_jwk(
            jwk=self.dict(exclude_none=True)
        )

    def is_asymmetric(self) -> bool:
        return self.kty in {'EC', 'OKP', 'RSA'}

    def is_loaded(self) -> bool:
        return True

    def is_public(self) -> bool:
        return all([
            getattr(self, attname) is None
            for attname in ['d', 'k', 'p', 'q', 'dp', 'dq', 'qi']
        ])

    def json(self, **kwargs: Any) -> str: # type: ignore pragma: no cover
        kwargs.setdefault('exclude_defaults', True)
        return super().json(**kwargs)



# Copyright 2022 Cochise Ruhulessin
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from typing import Any

import pydantic
from pyasn1.codec.der.encoder import encode as encode_der
from pyasn1.type.base import Asn1Type
from pyasn1.type.char import PrintableString
from pyasn1.type.char import UTF8String
from pyasn1_modules import rfc5280

from ckms.ext import asn1


class Attribute(pydantic.BaseModel):
    type: str
    value: Any

    @pydantic.validator('value', pre=True)
    def validate_value(
        cls,
        value: Any
    ) -> Any:
        obj = value
        if isinstance(value, rfc5280.AttributeValue):
            obj = asn1.decode_der(bytes(value))
        elif isinstance(value, Asn1Type):
            raise NotImplementedError(type(value))
        if isinstance(obj, (PrintableString, UTF8String)):
            obj = str(obj)
        return obj

    def asn1(self) -> rfc5280.AttributeTypeAndValue:
        obj = rfc5280.AttributeTypeAndValue()
        obj.setComponentByPosition(0, self.type) # type: ignore
        if not isinstance(self.value, str):
            raise NotImplementedError(type(self.value))
        obj.setComponentByPosition(1, encode_der(self.value, UTF8String())) # type: ignore
        return obj


# Copyright 2018 Cochise Ruhulessin
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from typing import cast
from typing import Any

import google.auth
import google.auth.transport.requests
import google.auth.exceptions
import pytest

from ckms.jose import TrustedIssuers
import pytest_asyncio


@pytest.fixture(scope='session')
def id_token() -> str:
    try:
        creds, _ = cast(tuple[Any, Any], google.auth.default()) # type: ignore
        creds.refresh(google.auth.transport.requests.Request())
    except google.auth.exceptions.RefreshError:
        pytest.skip()
    return creds.id_token


@pytest_asyncio.fixture(scope='session') # type: ignore
async def trust() -> TrustedIssuers:
    trust = TrustedIssuers()
    trust.trust("accounts.google.com")
    await trust
    return trust


@pytest.mark.parametrize("issuer", [
    "accounts.google.com",
    "https://accounts.google.com"
])
@pytest.mark.asyncio
async def test_trust(issuer: str):
    trust = TrustedIssuers()
    trust.trust(issuer=issuer)
    await trust


# Copyright 2022 Cochise Ruhulessin
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from typing import Any

import pydantic

from .localkey import LocalKey
from .localkeyspecification import LocalKeySpecification
from .transientkey import TransientKey


class Key(pydantic.BaseModel):
    __root__: LocalKey | TransientKey

    @classmethod
    def parse_obj(cls: type[pydantic.BaseModel], obj: Any) -> pydantic.BaseModel:
        return super().parse_obj(obj).__root__ # type: ignore

    async def setup(self, spec: LocalKeySpecification) -> 'Key':
        return await self.__root__.setup(spec)